"""
Third party WSGI Server setup helper function.

Mostly usefull during development and testing.
"""
__all__ = ['start_wsgi_server']

from _wsgi import build_handler

import logging
logger = logging.getLogger('fanery.server')


def start_wsgi_server(wsgi_app=None, host='127.0.0.1', port=9000, ssl=False,
                      bjoern=True, gevent=False, cherrypy=False, **argd):

    if wsgi_app is None:
        wsgi_app = build_handler(**argd)

    logger.info("serving on %s://%s:%s" % ('https' if ssl else 'http', host, port))     # noqa

    if ssl is True:
        from werkzeug.serving import run_simple

        params = dict(ssl_context=argd.get('ssl_context', 'adhoc'),
                      use_reloader=argd.get('use_reloader', True),
                      use_debugger=argd.get('use_debugger', True),
                      use_evalex=argd.get('use_evalex', True),
                      threaded=argd.get('threaded', True))

        try:
            run_simple(host, port, wsgi_app, **params)
        except KeyboardInterrupt:
            pass

    elif cherrypy is True:
        import cherrypy

        cherrypy.tree.graft(wsgi_app, '/')

        cherrypy.config.update({
            'engine.autoreload.on': True,
            'log.screen': True,
            'server.socket_port': port,
            'server.socket_host': host
        })

        try:
            cherrypy.engine.start()
            cherrypy.engine.block()
        except KeyboardInterrupt:
            cherrypy.engine.stop()

    elif gevent is True:
        from gevent.wsgi import WSGIServer

        try:
            WSGIServer((host, port), wsgi_app).serve_forever()
        except KeyboardInterrupt:
            pass

    elif bjoern is True:
        import bjoern

        try:
            bjoern.run(wsgi_app, host, port)
        except KeyboardInterrupt:
            pass

    else:
        from wsgiref.simple_server import make_server

        try:
            make_server(host, port, wsgi_app).serve_forever()
        except KeyboardInterrupt:
            pass



__all__ = ['PyClient', 'WsgiClient', 'HttpClient']

from base64 import b64encode, b64decode
from time import time as timestamp
from os.path import join
from cxor import xor
import scrypt

from _crypto import (
    nacl_nonce, nacl_random, nacl_box, nacl_sign, nacl_sha256, nacl_verify,
    Box, PrivateKey, PublicKey, VerifyKey, SigningKey,
)
from _term import Hict, parse_term, parse_json, to_simple, to_json
from _service import consume, prelogin, login, logout, gen_pad
import _exc as exc


class PyClient(object):
    """Python client

    The easy way to locally consume encrypted services from python programs.
    Mostly usefull during development and testing.
    """
    urlpaths = Hict(prelogin=prelogin.urlpath,
                    login=login.urlpath,
                    logout=logout.urlpath,
                    gen_pad=gen_pad.urlpath)
    password = Hict(hash_func=scrypt.hash)

    def __init__(self, **state):
        from _state import get_state
        self._state = get_state(**state)

    def _consume(self, urlpath, **argd):
        _, ext, ret = consume(self._state, urlpath, **argd)
        return ret

    def _build_urlpath(self, urlpath, args):
        if args:
            urlpath, ext = splitext(urlpath)
            urlpath = join(urlpath, *args) + ext
        return urlpath

    def login(self, login, password, force=False, **argd):
        urlpaths = self.urlpaths
        # prelogin
        ret = Hict(self._consume(urlpaths.prelogin, identity=login, **argd))
        password_hash = self.password.hash_func(password, b64decode(ret.salt))
        box, _, _ = nacl_box(b64decode(ret.pkey), nacl_sha256(password_hash))
        try:
            msg = box.decrypt(b64decode(ret.enc))
        except:
            raise exc.InvalidCredential(login, self._state)
        # login
        sid = b64encode(msg[64:80])
        cseed, vkey, pad = msg[:32], msg[32:64], msg[80:]
        sign, _ = nacl_sign(password_hash, cseed)
        ret = Hict(self._consume(urlpaths.login, identity=login, sid=sid,
                                 sign=b64encode(xor(sign[:64], pad)),
                                 force=force))
        msg = box.decrypt(nacl_verify(b64decode(ret.sign), vkey))
        self._auth = Hict(sid=sid,
                          tdiff=int(msg[64:]) - timestamp(),
                          sign_key=SigningKey(cseed),
                          verify_key=VerifyKey(vkey),
                          box=Box(PrivateKey(msg[:32]), PublicKey(msg[32:64])))
        return True

    def call(self, urlpath, *args, **argd):
        return self._consume(self._build_urlpath(urlpath, args), **argd)

    def safe_call(self, urlpath, *args, **argd):
        # get one-time-pad and one-time-keys
        auth = self._auth
        box = auth.box
        params = to_json(to_simple(argd))
        size = len(params) + 40
        pad = nacl_random(size + 340)
        msg = to_json(dict(call=urlpath, pad=b64encode(pad),
                           tstamp=int(timestamp()) + auth.tdiff))
        enc = box.encrypt(msg, nacl_nonce())
        sign = b64encode(auth.sign_key.sign(enc))
        ret = self._consume(self.urlpaths.gen_pad, sid=auth.sid, sign=sign)
        enc = auth.verify_key.verify(b64decode(ret['sign']))
        assert len(enc) < len(pad), (len(enc), len(pad))
        cpad = Hict(parse_json(box.decrypt(xor(enc, pad))))
        # build real request
        box, _, _ = nacl_box(b64decode(cpad.skey), b64decode(cpad.ckey))
        enc = xor(box.encrypt(params, nacl_nonce()), b64decode(cpad.pad))
        sign = b64encode(SigningKey(b64decode(cpad.cseed)).sign(enc))
        ret = self._consume(urlpath, sid=auth.sid, pid=cpad.pid, sign=sign)
        enc = VerifyKey(b64decode(cpad.vkey)).verify(b64decode(ret['sign']))
        return parse_json(box.decrypt(enc))

    def logout(self):
        return self.safe_call('logout')


from _wsgi import handler as wsgi_app
from os.path import splitext
from webob import Request
import requests


class WsgiClient(PyClient):

    def _parse_response(self, status, body):
        try:
            if status == 200:
                return parse_term(parse_json(body))
            else:
                error = Hict(parse_json(body))
        except:
            return body

        try:
            exc_class = (getattr(exc, error.exc, None) or
                         getattr(__builtins__, error.exc))
        except:
            return body

        raise exc_class(*error.err)

    def _consume(self, urlpath, **argd):
        params = dict((k, (v.name, v) if isinstance(v, file) else v)
                      for k, v in argd.iteritems()) if argd else None

        _, ext = splitext(urlpath)
        if not ext:
            urlpath = urlpath + '.json'

        if urlpath.endswith('.json'):
            params = dict(_json_=to_json(to_simple(params)))

        schema = 'https' if self._state.ssl is True else 'http'
        base_url = '%s://%s' % (schema, self._state.domain)
        req = Request.blank(urlpath, base_url=base_url, POST=params)

        res = req.get_response(wsgi_app)
        return self._parse_response(res.status_int, res.body)


class HttpClient(WsgiClient):

    def __init__(self, host='127.0.0.1', port=9000, **argd):
        super(HttpClient, self).__init__(**argd)
        scheme = 'http' if not self._state.ssl else 'https'
        self.__baseurl = '%s://%s:%s' % (scheme, host, port)

    def _consume(self, urlpath, **argd):
        url = '/'.join((self.__baseurl, urlpath))

        _, ext = splitext(url)
        if not ext:
            url = url + '.json'

        if argd:
            data, files = dict(), dict()

            for k, v in argd.iteritems():
                if isinstance(v, file):
                    files[k] = v
                else:
                    data[k] = v

            if url.endswith('.json'):
                data = dict(_json_=to_json(to_simple(data)))

            res = requests.post(url, files=files, data=data)

        else:
            res = requests.get(url)

        return self._parse_response(res.status_code, res.content)



from fanery._term import parse_term, Hict, UUID, is_uuid, is_dict, is_number
from fanery._random import rand_uuid, rand_num

# -- test helper functions --------------------------


def gen_dict():
    gen_ints = lambda: [str(rand_num(i, dec=False)) for i in xrange(1, 11)]
    gen_floats = lambda: [str(rand_num(i)) for i in xrange(1, 11)]
    gen_uuids = lambda: (str(rand_uuid()), rand_uuid().hex, rand_uuid())

    return dict(ints=gen_ints(),
                floats=gen_floats(),
                uuids=gen_uuids())

# -- real tests -------------------------------------


def test_parse_term_Hict():
    dct = gen_dict()
    dct.update(dicts=[gen_dict() for _ in range(10)])
    for d in dct['dicts']:
        d.update(dicts=[gen_dict() for _ in range(10)])

    assert is_dict(dct)

    term = parse_term(dct, dict_type=Hict)

    assert isinstance(term, Hict)
    assert is_dict(term)

    for u in term.uuids:
        assert isinstance(u, UUID)
        assert is_uuid(u)

    for i in term.ints:
        assert isinstance(i, (int, long))
        assert is_number(i)

    for f in term.floats:
        assert isinstance(f, (int, long, float))
        assert is_number(f)

    # recursive parse_term

    for dct in term.dicts:
        assert isinstance(dct, Hict)
        assert is_dict(dct)

        for u in dct.uuids:
            assert isinstance(u, UUID)
            assert is_uuid(u)

        for i in dct.ints:
            assert isinstance(i, (int, long))
            assert is_number(i)

        for f in dct.floats:
            assert isinstance(f, (int, long, float))
            assert is_number(f)

        for d in dct.dicts:
            assert isinstance(d, Hict)
            assert is_dict(d)

            for u in d.uuids:
                assert isinstance(u, UUID)
                assert is_uuid(u)

            for i in d.ints:
                assert isinstance(i, (int, long))
                assert is_number(i)

            for f in d.floats:
                assert isinstance(f, (int, long, float))
                assert is_number(f)

# -- tests run -------------------------------------------


def main():
    from random import shuffle
    tests = [(k, v) for k, v in globals().iteritems()
             if k.startswith('test_') and callable(v)]
    shuffle(tests)

    from fanery import timecall, memory_profile, line_profile  # noqa
    for name, test in tests:
        #memory_profile(test)()
        #line_profile(test)()
        timecall(test)()

if __name__ == '__main__':
    main()



# user session timeout in seconds
SESSION_TIMEOUT = 3600

# max active user sessions (negative means no limit)
MAX_ACTIVE_USER_SESSIONS = -1

# threshold of abuses by host to trigger rejection
MAX_ORIGIN_ABUSE_LEVEL = 10

# reset host abuses count after X seconds
ABUSE_LEVEL_WATCH_PERIOD = 600

# seconds granted to perform login after successfull prelogin
PRELOGIN_GRACE_TIME = 3

# max pending prelogin attempts by user
MAX_PRELOGIN_COUNT = 1

# seconds granted to perform safe call after successfull gen_pad
PAD_GRACE_TIME = 3

# max queued pads by user for the same safe call
MAX_PADS_COUNT = 1

# logging setup
LOGGING = {
    'version': 1,
    'disable_existing_loggers': True,

    'formatters': {
        'console': {
            'format': '%(asctime)s %(levelname)s %(name)s %(message)s',
        },
    },

    'handlers': {
        #'sentry': {
        #    'level': 'ERROR',
        #    'class': 'raven.handlers.logging.SentryHandler',
        #    'dsn': 'https://public:secret@example.com/1',
        #},
        #'graypy': {
        #    'level': 'INFO',
        #    'class': 'graypy.GELFRabbitHandler',
        #    'url': 'amqp://guest:guest@example.com:5672/%2F',
        #},
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'console',
        },
    },

    'loggers': {
        'fanery': {
            'handlers': ['console', ],  # 'sentry', 'graypy'],
            'level': 'DEBUG',
            'propagate': False,
        },
    },
}

# file upload folder
from tempfile import gettempdir
UPLOAD_DIRPATH = gettempdir()

# jfanery static files folder
from os.path import realpath, normpath, dirname, join
JFANERY_DIRPATH = join(realpath(normpath(dirname(__file__))), 'jfanery')
JFANERY_URLPATH = '/jfanery'



"""
Toy in-memory storage implementation.
"""

from collections import defaultdict
from time import time as timestamp
from _auth import hooks
from _term import Hict


class MemDictStore(object):

    def __init__(self):
        self._prelogins = defaultdict(dict)
        self._abuses = defaultdict(list)
        self._models = dict()
        self._sessions = dict()

    def do_nothing(self, *args, **argd):
        pass
    begin = commit = rollback = do_nothing

    def add_model(self, model, add_index):
        self._models[model] = dict()

    def store(self, domain, txn, inserts=None, updates=None, deletes=None):
        if inserts:
            self.upsert(domain, txn, inserts)
        if updates:
            self.upsert(domain, txn, updates)
        if deletes:
            self.delete(domain, txn, deletes)

    def upsert(self, domain, txn, records):
        for record in records:
            self._models[record._model][record._uuid] = record
    insert = update = upsert

    def delete(self, domain, txn, records):
        for record in records:
            del self._models[record._model][record._uuid]

    def fetch(self, domain, model, uuid, vsn=None, txn=None):
        record = self._models[model][uuid]
        assert not vsn or record._vsn == vsn
        assert not txn or record._txn == txn
        return record._vsn, record._txn, record._dct

    def query(self, domain, model, fun, score=None):
        for record in filter(fun, self._models[model].itervalues()):
            uuid, vsn, txn = record._uuid, record._vsn, record._txn
            if score is not None:
                yield (uuid, vsn, txn, score)
            else:
                yield (uuid, vsn, txn)

    def search(self, domain, model, fun):
        return self.query(domain, model, fun, 1)

    def select(self, domain, model, fun):
        return self.query(domain, model, fun)

    def find(self, domain, model, terms):
        for record in self._models[model].itervalues():
            if record.domain == domain:
                dct = record._dct
                for k, v in terms.iteritems():
                    if k in dct and dct[k] == v:
                        continue
                    else:
                        break
                else:
                    return record

    def fetch_profile(self, domain, identity, **extra):
        extra.update(username=identity)
        return self.find(domain, 'Profile', extra)

    def has_permission(self, state, call):
        return True

    def active_user_sessions(self, domain, identity):
        now = timestamp()
        return sum(s.expires > now
                   and s.domain == domain
                   and s.identity == identity
                   for s in self._sessions.itervalues())

    def destroy_sessions(self, domain, identity):
        sessions = self._sessions
        for sid, s in sessions.iteritems():
            if s.domain == domain and s.identity == identity:
                del sessions[sid]

    def load_state(self, domain, sid):
        return self._sessions.get(sid, None)

    def store_state(self, state):
        self._sessions[state.sid] = Hict(state)

    def destroy_state(self, state):
        del self._sessions[state.sid]

    def store_prelogin(self, origin, sid, **argd):
        self._prelogins[origin][sid] = Hict(argd)

    def fetch_prelogin(self, origin, domain, sid):
        prelogins = self._prelogins.get(origin, None)
        if prelogins and sid in prelogins:
            prelogin = prelogins.pop(sid)
            if prelogin.domain == domain:
                return prelogin

        return None

    def prelogin_count(self, domain, origin):
        tstamp = timestamp()
        prelogins = self._prelogins[origin]
        for k, v in prelogins.items():
            if v.expires <= tstamp:
                del prelogins[k]
        return len(prelogins)

    def abuse_level_by_origin(self, origin, domain):
        watch_period = timestamp() - hooks.abuse_level_watch_period(domain)
        return sum(l for t, l, _ in self._abuses[origin] if t >= watch_period)

    def incr_abuse(self, origin, domain, incr=1, abuse=None):
        self._abuses[origin].append((timestamp(), incr, abuse))



"""
Toy storage implementation using HyperDex.
"""

from _term import (
    Hict, parse_term, is_number,
    dump_term as dumps, load_term as loads
)
from _auth import default_settings, hooks
from _random import rand_uuid
from _state import get_state
from _admin import Profile
from _store import Record
from _exc import RecordVersionMismatch, DuplicatedRecordID

import _config as conf

from time import time as timestamp
from bsdiff4 import diff as bsdiff
from hyperdex.admin import Admin
from hyperdex.client import Client

# default models names
Profile = Profile.__name__
State = '_state_'
Txn = '_txn_'
Prelogin = '_prelogin_'
Abuse = '_abuse_'
Acl = '_acl_'
Settings = '_settings_'

# schema templates
spaces = Hict(index='''
              space _idx_%s
              key id
              attributes
                string domain,
                string uuid,
                int vsn,
                string field,
                int num,
                float dec,
                string str
              subspace domain, uuid
              subspace domain, field, num
              subspace domain, field, dec
              subspace domain, field, str
              ''',
              model='''
              space %s
              key uuid
              attributes
                string domain,
                int vsn,
                string txn,
                string data
              subspace domain, txn
              ''',
              hist='''
              space %s_hist
              key id
              attributes
                string domain,
                string uuid,
                int vsn,
                string txn,
                string data
              subspace domain, uuid, vsn
              subspace domain, txn
              ''',
              base=('''
                    space %s
                    key txn
                    attributes
                        string domain,
                        string origin,
                        string profile,
                        float tstamp,
                        string data
                    subspace domain, profile
                    subspace domain, tstamp
                        ''' % Txn, '''
                    space %s
                    key sid
                    attributes
                        string domain,
                        string origin,
                        string identity,
                        string profile,
                        float expires,
                        string data
                    subspace domain, origin, expires
                    subspace domain, identity
                        ''' % State, '''
                    space %s
                    key sid
                    attributes
                        string domain,
                        string origin,
                        string identity,
                        float expires,
                        string data
                    subspace domain, origin, expires
                    subspace domain, identity, expires
                        ''' % Prelogin, '''
                    space %s
                    key id
                    attributes
                        string origin,
                        float tstamp,
                        int level,
                        string data
                    subspace origin, tstamp
                        ''' % Abuse, '''
                    space %s
                    key domain
                    attributes
                        map(string,string) settings
                        ''' % Settings))


class HDStore(object):

    def __init__(self, host='localhost', port=1982):
        self.adm = adm = Admin(host, port)
        self.cli = Client(host, port)
        self.indexed = set()

        # initialize fanery base spaces
        for space in spaces.base:
            try:
                adm.add_space(space)
            except:
                # do nothing, exists already
                pass

    # -- model spaces ----------------------------------------

    def add_model(self, model, add_index=False):
        model_spaces = [spaces.model, spaces.hist]
        if add_index:
            model_spaces.append(spaces.index)
            self.indexed.add(model)

        adm = self.adm
        for space in model_spaces:
            try:
                adm.add_space(space % model)
            except:
                pass

    # -- transaction -----------------------------------------

    def begin(self, domain, txn):
        pass

    def commit(self, domain, txn, state, log=None):
        profile = str(state.profile._uuid) if state.profile else None
        data = dumps(dict(log=log, state=state or get_state()))
        self.cli.put(Txn, str(txn), dict(domain=domain, origin=state.origin,
                                         tstamp=timestamp(), profile=profile,
                                         data=data))

    def rollback(self, domain, txn):
        pass

    # -- crud ------------------------------------------------

    def fetch(self, domain, model, uuid, vsn=None, txn=None):
        qry = dict(domain=domain, uuid=uuid)
        if vsn is not None:
            qry['vsn'] = vsn
        if txn is not None:
            qry['txn'] = txn

        for row in self.cli.search(model, qry):
            return (row['vsn'], row['txn'], loads(row['data']))

    def store(self, domain, txn, inserts=None, updates=None, deletes=None):
        if inserts:
            self.insert(domain, txn, inserts)
        if updates:
            self.update(domain, txn, updates)
        if deletes:
            self.delete(domain, txn, deletes)

    #TODO: use async insert/update/delete?

    def insert(self, domain, txn, records):
        indexed = self.indexed
        inserts = set()
        cli = self.cli

        for record in records:
            model, uuid = record._model, str(record._uuid)

            data = dict(domain=domain, vsn=record._vsn,
                        txn=str(txn), data=dumps(record._dct))

            if not cli.put_if_not_exist(model, uuid, data):
                raise DuplicatedRecordID(domain, model, uuid)
            elif model in indexed:
                inserts.add(record)

        if inserts:
            self.index(domain, txn, inserts=inserts)

    def update(self, domain, txn, records):
        indexed = self.indexed
        updates = set()
        cli = self.cli

        for record in records:
            model, uuid, vsn = record._model, str(record._uuid), record._vsn
            stored = cli.get(model, uuid)
            data = dumps(record._dct)
            old_vsn = stored['vsn']

            if not ((old_vsn == vsn - 1) and
                    cli.cond_put(model, uuid, dict(vsn=vsn - 1),
                                 dict(domain=domain, vsn=vsn, txn=str(txn),
                                      data=data))):
                raise RecordVersionMismatch(domain, model, uuid, vsn)
            else:
                # compute reverse binary diff
                stored['data'] = bsdiff(data, stored['data'])
                cli.put('%s_hist' % model, '%s:%s' % (uuid, old_vsn), stored)

                if model in indexed:
                    updates.add(record)

        if updates:
            self.index(domain, txn, updates=updates)

    def delete(self, domain, txn, records):
        indexed = self.indexed
        deletes = set()
        cli = self.cli

        for record in records:
            model, uuid, vsn = record._model, str(record._uuid), record._vsn
            stored = cli.get(model, uuid)

            if not ((stored['vsn'] == -vsn) and
                    cli.cond_put(model, uuid, dict(vsn=-vsn),
                                 dict(domain=domain, vsn=vsn, txn=str(txn)))):
                raise RecordVersionMismatch(domain, model, uuid, vsn)
            else:
                cli.put('%s_hist' % model, '%s:%s' % (uuid, -vsn), stored)

                if model in indexed:
                    deletes.add(record)

        if deletes:
            self.index(domain, txn, deletes=deletes)

    def index(self, domain, txn, inserts=None, updates=None, deletes=None):
        purge_record_index = self.cli.delete
        store_index = self.cli.put_if_not_exist

        for put, records in ((True, inserts), (True, updates), (False, deletes)):   # noqa
            if not records:
                continue

            for record in records:
                model, uuid = '_idx_%s' % record._model, str(record._uuid)

                purge_record_index(model, dict(domain=domain, uuid=uuid))

                if put:
                    #FIXME: handle date/datetime terms
                    for field, term in record._indexer(record).items():
                        data = dict(domain=domain, uuid=uuid, vsn=record._vsn)
                        if isinstance(term, (int, long)):
                            data.update(field=field, num=term)
                        elif is_number(term):
                            data.update(field=field, dec=term)
                        elif isinstance(term, basestring):
                            data.update(field=field, str=term)
                        else:
                            continue
                        while not store_index(model, str(rand_uuid()), data):
                            continue

    # -- fuzzy, relational and by-terms query --------------------

    def search(self, domain, model, fun):
        raise NotImplementedError

    def select(self, domain, model, fun):
        yield fun(self.cli, domain, model)

    def find(self, domain, model, terms):
        model = '_idx_%s' % model

        qry = dict(domain=domain)
        #FIXME: handle date/datetime terms
        #FIXME: handle range criteria
        for k, v in terms.items():
            if isinstance(v, (int, long)):
                qry.update(field=k, num=v)
            elif is_number(v):
                qry.update(field=k, dec=v)
            elif isinstance(v, basestring):
                qry.update(field=k, str=str(v))

        for dct in self.cli.search(model, qry):
            if dct['domain'] == domain:
                yield dct['uuid'], dct['vsn']

    # -- auth hooks --------------------------------------------

    def _fetch_setting(self, domain, key, fail=None, parser=parse_term):
        try:
            settings = self.cli.get(Settings, domain)['settings']
        except:
            settings = dict((k, str(getattr(conf, k.upper())))
                            for k in default_settings)
            assert self.cli.put(Settings, domain, dict(settings=settings))
        if key not in settings:
            return fail
        elif parser is not None:
            return parser(settings[key])
        else:
            return settings[key]

    def _fetch_record(self, domain, model, uuid=None, **terms):
        if uuid:
            stored = self.cli.get(model, str(uuid))
            vsn = stored['vsn']
        elif terms:
            for uuid, vsn in self.find(domain, model, terms):
                stored = self.cli.get(model, uuid)
                if stored['vsn'] == vsn:
                    break

        if stored['domain'] == domain:
            return Record(model, uuid=uuid, vsn=vsn, txn=stored['txn'],
                          **loads(stored['data']))

    def fetch_profile(self, domain, identity, **extra):
        return self._fetch_record(domain, Profile, username=identity, **extra)

    def prelogin_count(self, domain, origin):
        return self.cli.count(Prelogin, dict(domain=domain, origin=origin))

    def fetch_prelogin(self, origin, domain, sid):
        cli = self.cli
        try:
            prelogin = Hict(cli.get(Prelogin, sid))
            if prelogin.domain == domain and prelogin.origin == origin:
                cli.delete(Prelogin, sid)
                prelogin.update(**loads(prelogin.pop('data')))
                return prelogin
        except:
            pass

    def store_prelogin(self, origin, domain, sid, identity, expires, **data):
        self.cli.put(Prelogin, sid, dict(origin=origin, domain=domain,
                                         identity=str(identity),
                                         expires=expires,
                                         data=dumps(data)))

    def load_state(self, domain, sid):
        try:
            stored = self.cli.get(State, sid)
            if stored['domain'] == domain:
                profile = self._fetch_record(domain, Profile, stored.pop('profile'))  # noqa
                state = loads(stored.pop('data'), dict_type=Hict)
                state.update(sid=sid, profile=profile, **stored)
                return state
        except:
            pass

    def store_state(self, state):
        profile = state.profile
        data = dict(skey=state.skey, ckey=state.ckey, sseed=state.sseed,
                    vkey=state.vkey, pads=state.pads, data=state.data,
                    role=state.role, impersonated_by=state.impersonated_by)
        assert self.cli.put(State, state.sid,
                            dict(identity=str(profile.username),
                                 profile=str(profile._uuid),
                                 domain=state.domain,
                                 origin=state.origin,
                                 data=dumps(data),
                                 expires=state.expires))

    def destroy_state(self, state):
        self.cli.delete(State, state.sid)

    def active_user_sessions(self, domain, identity):
        return self.cli.count(State, dict(domain=domain, identity=identity,
                                          expires=(timestamp(), 2 ** 31)))

    def destroy_user_sessions(self, domain, identity):
        self.cli.group_del(State, dict(domain=domain, identity=identity))

    def incr_abuse(self, origin, domain, incr=1, abuse=None):
        self.cli.put(Abuse, str(rand_uuid()),
                     dict(origin=origin, level=incr, tstamp=timestamp(),
                          data=dumps(abuse) if abuse is not None else abuse))

    def abuse_level_by_origin(self, origin, domain):
        watch_period = timestamp() - hooks.abuse_level_watch_period(domain)
        query = dict(origin=origin, tstamp=(watch_period, float(2 ** 31)))
        return sum(doc['level'] for doc in self.cli.search(Abuse, query))

    def reset_origin_abuse_level(self, origin, domain):
        self.cli.group_del(Abuse, dict(origin=origin))

    def session_timeout(self, domain):
        return self._fetch_setting(domain, 'session_timeout')

    def max_prelogin_count(self, domain):
        return self._fetch_setting(domain, 'max_prelogin_count')

    def abuse_level_watch_period(self, domain):
        return self._fetch_setting(domain, 'abuse_level_watch_period')

    def max_origin_abuse_level(self, domain):
        return self._fetch_setting(domain, 'max_origin_abuse_level')

    def max_active_user_sessions(self, domain, identity):
        return self._fetch_setting(domain, 'max_active_user_sessions')

    def has_permission(self, state, call):
        return True



__all__ = ['service', 'static']

from _term import (
    Hict, parse_term, parse_json, to_simple, to_json, to_str, is_str
)
from _crypto import (
    nacl_box, nacl_sign, nacl_verify,
    nacl_nonce, nacl_random, nacl_sha256,
    PrivateKey, PublicKey, Box, SigningKey
)
from _auth import hooks as auth
from _state import get_state
import _config as conf
import _exc as exc

from os.path import splitext, normpath, realpath, isdir, isfile, join
from base64 import b64encode, b64decode
from time import time as timestamp
from functools import wraps
from sys import exc_info
from cxor import xor

import logging
logger = logging.getLogger('fanery.service')

try:
    import ipdb as pdb
except ImportError:
    import pdb

try:
    import xtraceback
    traceback = xtraceback.compat
    traceback.install_traceback()
    traceback.install_sys_excepthook()
except ImportError:
    import traceback

# services registry -> see service() and lookup()
registry = dict()


def is_abusive(domain, origin):
    max_abuse_level = auth.max_origin_abuse_level(domain)
    if max_abuse_level > 0:
        origin_abuse_level = auth.abuse_level_by_origin(origin, domain)
        return bool(origin_abuse_level >= max_abuse_level)


def decrypt(keys, sign, pad=None):
    enc = nacl_verify(sign, keys.vkey)
    box, _, _ = nacl_box(keys.ckey, keys.skey)
    msg = box.decrypt(xor(enc, pad) if pad else enc)
    return box, parse_json(msg)


def encrypt(state, box, term, pad=None):
    enc = box.encrypt(to_json(to_simple(term)), nacl_nonce())
    sign = nacl_sign(xor(enc, pad) if pad else enc, state.sseed)[0]
    return Hict(sign=b64encode(sign))


def service(urlpath=None, auto_parse=True, static=False, accel=False,
            cache=False, force_download=False, content_disposition=False,
            auto_close_file=True, safe=True, ssl=True, use_one_time_pad=True,
            check_abusive=True, check_perm=True, postmortem_debug=False,
            force_define=False, log_exc=True, **output):

    for k, v in output.iteritems():
        if not (k and k.isalnum() and callable(v)):
            raise exc.InvalidOutputFormatter(k, v)
    else:
        output.setdefault('json', to_json)
        output.setdefault('txt', to_str)
        output.setdefault('raw', repr)

    if not safe:
        use_one_time_pad = False

    def decorator(f):
        srv_path = (urlpath or f.__name__).strip('/')

        if srv_path in registry and not force_define:
            raise exc.MultipleDefine(srv_path)

        @wraps(f)
        def wrapper(*args, **argd):
            _state_ = get_state()

            if ssl is True and not _state_.ssl:
                raise exc.RequireSSL(srv_path)

            domain = _state_.domain
            origin = _state_.origin

            if check_abusive and is_abusive(domain, origin):
                raise exc.Abusive(origin)

            if safe is True:

                sid = b64decode(argd.pop('sid', ''))
                if not sid:
                    auth.incr_abuse(origin, domain,
                                    abuse=('sid', None, srv_path))
                    raise exc.Unauthorized

                stored = auth.load_state(domain, sid)
                if not stored:
                    auth.incr_abuse(origin, domain,
                                    abuse=('sid', sid, srv_path))
                    raise exc.UnknownSession
                elif stored.expires <= timestamp():
                    raise exc.ExpiredSession
                else:
                    _state_.update(stored)
                    _state_.origin = origin

                if check_perm and not auth.has_permission(_state_, srv_path):
                    auth.incr_abuse(origin, domain, incr=3,
                                    abuse=('auth', sid, srv_path))
                    raise exc.Unauthorized

                sign = b64decode(argd.pop('sign', ''))
                if not sign:
                    auth.incr_abuse(origin, domain, incr=3,
                                    abuse=('sign', sid, srv_path))
                    raise exc.InvalidCall('sign')

                if use_one_time_pad is True:
                    pid, pads = b64decode(argd.pop('pid', '')), stored.pads
                    cpads = pads.get(srv_path, None)
                    if not pid or not cpads or pid not in cpads:
                        auth.incr_abuse(origin, domain, incr=3,
                                        abuse=('pid', sid, srv_path))
                        raise exc.InvalidCall('pid')

                    keys = cpads.pop(pid)
                    pad = keys.pad
                    if not cpads:
                        del pads[srv_path]
                else:
                    keys = _state_
                    pad = None

                try:
                    box, term = decrypt(keys, sign, pad)
                except:
                    auth.incr_abuse(origin, domain, incr=3,
                                    abuse=('pad', sid, srv_path))
                    raise exc.InvalidCall('pad')

                argd.update(term)

            if auto_parse is True:
                args = parse_term(args) if args else args
                argd = parse_term(argd) if argd else argd

            try:
                _state_.service = srv_path

                if not use_one_time_pad and safe is True:
                    ret = f(box, *args, **argd)
                else:
                    ret = f(*args, **argd)

                if safe is True:
                    if _state_.expires > 0:
                        _state_.expires = timestamp() + auth.session_timeout(domain)    # noqa
                        auth.store_state(_state_)
                    else:
                        auth.destroy_state(_state_)

            except Exception:

                exc_type, exc_value, exc_traceback = exc_info()

                if postmortem_debug is True:
                    traceback.print_exc()
                    pdb.post_mortem(exc_traceback)
                elif log_exc is True:
                    logger.error(srv_path, exc_info=True)

                if safe is True:
                    ret = dict(low=isinstance(exc_type, (AssertionError, exc.FaneryException)), # noqa
                               exc=exc_type.__name__, err=to_simple(exc_value.args))            # noqa
                    _state_.error.update(ret)
                else:
                    raise

            if use_one_time_pad is True:
                return encrypt(keys, box, ret)
            else:
                return ret

        wrapper.ssl = ssl
        wrapper.safe = safe
        wrapper.cache = cache
        wrapper.static = static
        wrapper.output = output
        wrapper.urlpath = srv_path
        wrapper.auto_parse = auto_parse
        wrapper.force_download = force_download
        wrapper.auto_close_file = auto_close_file
        wrapper.content_disposition = content_disposition
        wrapper.accel = accel.lower() if is_str(accel) else None

        registry[srv_path] = wrapper

        return wrapper

    return decorator


def lookup(urlpath, full=None):

    path, ext = splitext(urlpath.strip().strip('/'))
    args = []

    while path not in registry:
        try:
            path, arg = path.rsplit('/', 1)
            args.append(arg)
        except ValueError:
            break

    fun = registry.get(path, None)

    if not fun:
        try:
            prefix, urlpath = urlpath.split('/', 1)
            return lookup(urlpath, full or urlpath)
        except ValueError:
            fun = registry.get('', None)
            args = [full]
    elif args:
        args.reverse()
        if fun and fun.static is True:
            args[-1] += ext

    out = fun.output.get(ext[1:], False) if ext and fun else None

    return fun, args, ext, out


def consume(_state_, urlpath, *args, **argd):
    start = timestamp()
    try:
        fun = error = exception = None
        fun, argv, ext, out = lookup(urlpath)

        if not callable(fun):
            raise exc.NotFound(urlpath)

        if ext and not (fun.static or callable(out)):
            raise exc.UndefinedOutputFormatter(urlpath)

        argv.extend(args)
        ret = fun(*argv, **argd)

        return fun, ext, out(ret) if ext and out else ret
    except exc.FaneryException, error:
        raise
    except Exception, exception:
        raise
    finally:
        sid, profile = _state_.sid, _state_.profile
        msg = "%s %s %s %s %s %s %s %s %0.6f %s" % (
            'S' if _state_.ssl else '-',
            'A' if fun and fun.auto_parse else '-',
            'S' if fun and fun.static else '-',
            _state_.origin, _state_.domain,
            sid.encode('hex') if sid else '-',
            profile.username if profile else '-',
            _state_.role or '-',
            timestamp() - start,
            urlpath)

        if _state_.error:
            logger.error("%s [%s]" % (msg, _state_.error.exc))
        elif error:
            logger.warning("%s [%s]" % (msg, error.__class__.__name__))
        elif exception:
            logger.error("%s [%s]" % (msg, exception.__class__.__name__))
        else:
            logger.info(msg)


def static(urlpath, root, index='index.html', **argd):

    root = join(realpath(normpath(root)), '')
    if not isdir(root):
        raise exc.NotFound(root)

    argd.setdefault('cache', True)
    argd.setdefault('log_exc', False)

    @service(urlpath, auto_parse=False, static=True, safe=False, **argd)
    def serve_file(*args, **argd):
        filepath = realpath(normpath(join(root, *args)))

        if isdir(filepath):
            if index:
                filepath = join(filepath, index)
            elif filepath.startswith(root):
                return filepath
            else:
                raise exc.NotFound(join(*args or [index]))

        if isfile(filepath) and filepath.startswith(root):
            return filepath
        else:
            raise exc.NotFound(join(*args or [index]))

    return serve_file

static(conf.JFANERY_URLPATH, conf.JFANERY_DIRPATH)


@service(ssl=False, safe=False, auto_parse=False, log_exc=False)
def prelogin(identity, **extra):
    state = get_state()
    origin = state.origin
    domain = state.domain

    max_count = auth.max_prelogin_count(domain)
    prelogin_count = auth.prelogin_count(domain, origin)
    if max_count <= 0 or prelogin_count < max_count:

        profile = auth.fetch_profile(domain, identity, **extra)
        if profile is not None:
            sid = auth.gen_session_id()
            hash = profile.password_hash
            ckey = PrivateKey(nacl_sha256(hash))
            skey = PrivateKey.generate()
            sseed = nacl_random(32)
            cseed = nacl_random(32)

            box = Box(skey, ckey.public_key)
            csign = SigningKey(cseed)
            sign = csign.sign(hash)[:64]
            pad = nacl_random(len(sign))
            sign = xor(sign, pad)

            auth.store_prelogin(expires=timestamp() + auth.prelogin_grace_time(domain),         # noqa
                                sid=sid, domain=domain, origin=origin,
                                identity=identity, sign=sign,
                                profile=profile._uuid,
                                ckey=ckey.public_key.encode(),
                                vkey=csign.verify_key.encode(),
                                skey=skey.encode(), sseed=sseed)

            vkey = SigningKey(sseed).verify_key.encode()
            enc = box.encrypt(cseed + vkey + sid + pad, nacl_nonce())
            return Hict(pkey=b64encode(skey.public_key.encode()),
                        salt=b64encode(profile.password_salt),
                        enc=b64encode(enc))
        else:
            auth.incr_abuse(origin, domain, abuse=('prelogin', identity))

    # do not tell directly prelogin attempt was unsuccessful
    # let the client waste time trying to decrypt some garbage
    return Hict(pkey=b64encode(PrivateKey.generate().public_key.encode()),
                salt=b64encode(nacl_random(8)),
                enc=b64encode(nacl_random(184)))


@service(ssl=False, safe=False, auto_parse=False, log_exc=False)
def login(identity, sid, sign, force=False):
    state = get_state()
    origin = state.origin
    domain = state.domain

    sign = b64decode(sign)
    sid = b64decode(sid)

    prelogin = auth.fetch_prelogin(origin, domain, sid)
    profile = auth.fetch_profile(domain, identity) if prelogin else None

    if not profile or not prelogin or timestamp() >= prelogin.expires:
        auth.incr_abuse(origin, domain, abuse=('login', identity))
    elif prelogin.profile != profile._uuid or prelogin.sign != sign:
        auth.incr_abuse(origin, domain, incr=3, abuse=('sign', identity))
    else:
        max_sessions = auth.max_active_user_sessions(domain, identity)
        if max_sessions > 0:
            active_sessions = auth.active_user_sessions(domain, identity)
            if not force and active_sessions >= max_sessions:
                raise exc.MaxActiveSessions
            elif force:
                auth.destroy_user_sessions(domain, identity)

        box = Box(PrivateKey(prelogin.skey), PublicKey(prelogin.ckey))
        skey = PrivateKey.generate()
        ckey = PrivateKey.generate()
        expires = timestamp() + auth.session_timeout(domain)

        state.update(sid=sid, profile=profile, expires=expires,
                     ckey=ckey.public_key.encode(), skey=skey.encode(),
                     sseed=prelogin.sseed, vkey=prelogin.vkey)
        auth.store_state(state)

        tstamp = str(int(timestamp()))
        msg = ckey.encode() + skey.public_key.encode() + tstamp
        enc = box.encrypt(msg, nacl_nonce())
        sign = SigningKey(prelogin.sseed).sign(enc)
        return Hict(sign=b64encode(sign))

    raise exc.InvalidCredential


@service(ssl=False, use_one_time_pad=False, auto_parse=False, log_exc=False)
def gen_pad(box, call, pad, tstamp):
    state = get_state()

    if not state.profile:
        raise exc.Unauthorized

    try:
        tsdiff = timestamp() - float(tstamp)
    except:
        raise exc.InvalidCall

    domain = state.domain
    pads = state.pads

    if abs(tsdiff) > auth.pad_grace_time(domain, call):
        auth.incr_abuse(state.origin, domain, abuse=('timeout', call))
        raise exc.CallPadTimeout(dict(tsdiff=tsdiff))
    elif not auth.has_permission(state, call):
        auth.incr_abuse(state.origin, domain, abuse=('denied', call))
        raise exc.Unauthorized
    elif call in pads and len(pads[call]) >= auth.max_pads_count(domain, call):
        auth.incr_abuse(state.origin, domain, abuse=('pads', call))
        raise exc.CallPadsLimitExceeded

    # gen one-time-pad and one-time-keys
    cid = nacl_random(16)
    cseed = nacl_random(32)
    sseed = nacl_random(32)
    skey = PrivateKey.generate()
    ckey = PrivateKey.generate()

    # server call pad data
    pad = b64decode(pad)
    size = len(pad) - 340
    cvkey = SigningKey(cseed).verify_key
    cpad = pads[call][cid] = Hict(pad=nacl_random(size), call=call,
                                  skey=skey.encode(), sseed=sseed,
                                  ckey=ckey.public_key.encode(),
                                  vkey=cvkey.encode())

    # client pad data
    svkey = SigningKey(sseed).verify_key
    data = Hict(pad=b64encode(cpad.pad),
                pid=b64encode(cid),
                cseed=b64encode(cseed),
                ckey=b64encode(ckey.encode()),
                skey=b64encode(skey.public_key.encode()),
                vkey=b64encode(svkey.encode()))

    return encrypt(state, box, data, pad)


@service(ssl=False, auto_parse=False, log_exc=False)
def logout(*args, **argd):
    _state_ = get_state()
    _state_.expires = 0
    return True



"""
Profiles model reference implementation.
"""

from time import time as timestamp

import logging
logger = logging.getLogger('fanery.admin')


class Profile:

    @classmethod
    def initialize(cls, record):
        from _state import get_state
        record.merge(domain=get_state().domain)

    @classmethod
    def validate(cls, record):
        from _term import is_str, is_email, Hict

        errors, dct = Hict(), record._dct

        domain = dct.get('domain')
        if not domain:
            errors.domain.required = domain

        username = dct.get('username')
        if not username:
            errors.username.required = username
        elif not is_str(username):
            errors.username.bad_type = type(username).__name__

        password_hash = dct.get('password_hash')
        if not password_hash:
            errors.password.required = password_hash

        email = dct.get('email')
        if not email:
            errors.email.required = email
        elif not is_str(email):
            errors.email.bad_type = type(email).__name__
        elif not is_email(email):
            errors.email.bad_format = email

        if errors:
            print dct
        return errors

    @classmethod
    def index(cls, record):
        return dict((k, v) for k, v in record._dct.iteritems()
                    if not k.startswith('password_'))


def setup(db):
    from _store import add_model, add_storage

    start = timestamp()
    try:
        add_model(Profile)
        add_storage(db, Profile)
    finally:
        logger.debug('%0.6f setup' % (timestamp() - start))


def init(admin_username='admin', admin_password=None, **extra):
    from _crypto import nacl_random
    from _state import get_state
    from _store import storage
    from _term import is_str
    from _auth import hooks
    from scrypt import hash

    start = timestamp()
    try:
        state = get_state()
        domain = state.domain = extra.setdefault('domain', state.domain)
        extra.setdefault('email', '%s@%s' % (admin_username, 'example.com'))

        admin = hooks.fetch_profile(domain, admin_username)
        if admin is None:
            if not admin_password:
                admin_password = nacl_random(16).encode('hex')
            else:
                assert is_str(admin_password), 'bad-type'

            salt = nacl_random(16)
            password_hash = hash(admin_password, salt)

            desc = "adduser %s:%s@%s" % (admin_username, admin_password, domain)    # noqa
            with storage(desc) as db:
                db.insert(Profile,
                          username=admin_username,
                          password_hash=password_hash,
                          password_salt=salt,
                          **extra)
    finally:
        logger.debug('%0.6f init' % (timestamp() - start))



__all__ = ['acl', 'hooks']

from functools import wraps

from _term import Hict, gen_uuid_bytes
from _exc import Unauthorized
import _config as conf

acl_rules = dict(role=lambda s, roles, *args, **argd: s.role in roles and role in s.profile.roles,  # noqa
                 user=lambda s, users, *args, **argd: s.profile.username in users,                  # noqa
                 origin=lambda s, origins, *args, **argd: s.origin in origins,
                 domain=lambda s, domains, *args, **argd: s.domain in domains)


def set_acl_rule(name, func):
    acl_rules[name] = func


def acl(**rules):

    for k, v in rules.iteritems():
        validator = acl_rules.get(k, v)
        assert callable(validator), ('invalid-acl', k)
        rules[k] = (v, validator)

    def decorator(f):

        from _state import get_state

        @wraps(f)
        def wrapper(*args, **argd):
            _state_ = get_state()
            for k, (v, validator) in rules.iteritems():
                try:
                    if v is validator:
                        valid = validator(_state_, *args, **argd)
                    else:
                        valid = validator(_state_, v, *args, **argd)
                except:
                    valid = False

                if valid is True:
                    continue
                elif valid is False:
                    raise Unauthorized(k)
                else:
                    raise Unauthorized(k, valid)

            return f(*args, **argd)

        return wrapper

    return decorator


def _undefined(name):

    def wrapper(*argv, **argd):
        raise NotImplementedError(name)

    return wrapper


def _default(key):
    key = key.upper()

    assert hasattr(conf, key), key

    def _default(*args, **argd):
        return getattr(conf, key)
    _default.is_default = True

    return _default

hooks = Hict(load_state=_undefined('hooks.load_state'),
             store_state=_undefined('hooks.store_state'),
             active_user_sessions=_undefined('hooks.active_user_sessions'),
             destroy_user_sessions=_undefined('hooks.destroy_user_sessions'),
             max_active_user_sessions=_default('max_active_user_sessions'),
             fetch_profile=_undefined('hooks.fetch_profile'),
             has_permission=_undefined('hooks.has_permission'),
             store_prelogin=_undefined('hooks.store_prelogin'),
             incr_abuse=_undefined('hooks.incr_abuse'),
             max_pads_count=_default('max_pads_count'),
             pad_grace_time=_default('pad_grace_time'),
             max_origin_abuse_level=_default('max_origin_abuse_level'),
             abuse_level_watch_period=_default('abuse_level_watch_period'),
             abuse_level_by_origin=_undefined('hooks.abuse_level_by_origin'),
             prelogin_grace_time=_default('prelogin_grace_time'),
             prelogin_count=_undefined('hooks.prelogin_count'),
             max_prelogin_count=_default('max_prelogin_count'),
             session_timeout=_default('session_timeout'),
             gen_session_id=gen_uuid_bytes)

default_settings = tuple(k for k, v in hooks.items()
                         if getattr(v, 'is_default', False))



from inspect import (  # noqa
    ismodule as is_module,
    isfunction as is_function,
    isgenerator as is_generator,
    isbuiltin as is_builtin,
    isclass as is_class
)

from datetime import datetime, date
from collections import Iterable
from decimal import Decimal
from uuid import UUID
from os import path


class Hict(dict):
    """
    Hierarchical dotted dictionary.
    """

    def __missing__(self, key):
        term = self[key] = Hict()
        return term

    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def is_str(term):
    return isinstance(term, basestring)


def is_number(term):
    return isinstance(term, (int, long, float, Decimal))


def is_date(term):
    return isinstance(term, (date, datetime))


def is_uuid(term):
    return isinstance(term, UUID)


def is_sequence(term):
    return hasattr(term, '__iter__') \
        and not isinstance(term, (basestring, dict)) \
        and not is_generator(term)


def is_dict(term):
    return isinstance(term, dict) or type(term) is dict


def is_inet_address(term):
    raise NotImplementedError


def is_inet6_address(term):
    raise NotImplementedError


def is_file_path(term):
    try:
        return path.isfile(term)
    except:
        return False


def is_dir_path(term):
    try:
        return path.isdir(term)
    except:
        return False


try:
    from libuuid import uuid4 as gen_uuid, uuid4_bytes as gen_uuid_bytes   # noqa
except ImportError:
    from uuid import uuid4 as gen_uuid      # noqa

    def gen_uuid_bytes():
        return gen_uuid().bytes


from re import compile as regex
try:
    from validate_email import validate_email as is_email
except ImportError:
    from email.utils import parseaddr as _parse_email_addr
    # borrowed from http://www.regular-expressions.info/email.html
    _email_regex = regex(r'''[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*@(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?''')  # noqa

    def is_email(term, verify=False):
        try:
            name, email = _parse_email_addr(term)
            return _email_regex.match(email)
        except:
            return False


def to_str(term):
    #FIXME: handle any term
    return '%s' % term


def to_simple(term):
    if not term or isinstance(term, (int, float, basestring)):
        return term
    elif isinstance(term, dict) or type(term) is dict:
        return dict((to_simple(k), to_simple(v))
                    for k, v in term.iteritems())
    elif isinstance(term, Iterable):
        return [to_simple(t) for t in term]
    else:
        return str(term)


try:
    from ciso8601 import parse_datetime
except ImportError:
    try:
        import arrow

        def parse_datetime(term):
            return arrow.get(term).datetime
    except ImportError:
        def parse_datetime(term):
            return datetime.strptime(term, "%Y-%m-%dT%H:%M:%S.%f")


def parse_term(term, dict_type=dict, parse_dict_key=False):
    if term and isinstance(term, basestring):
        try:
            f = float(term)
            i = int(f)
            if i == f:
                return i
            elif str(f) == term:
                return f
            else:
                return Decimal(term)
        except:
            pass
        try:
            return UUID(term)
        except:
            pass
        try:
            dt = parse_datetime(term)
            return dt if dt.time() else dt.date()
        except:
            return term
    elif isinstance(term, dict) or type(term) is dict:
        if parse_dict_key is True:
            return dict_type((parse_term(k, dict_type, parse_dict_key),
                              parse_term(v, dict_type, parse_dict_key))
                             for k, v in term.iteritems())
        else:
            return dict_type((k, parse_term(v, dict_type, parse_dict_key))
                             for k, v in term.iteritems())
    elif is_sequence(term):
        return type(term)(parse_term(t, dict_type, parse_dict_key) for t in term)   # noqa
    else:
        return term


try:
    from msgpack import packb as to_msgpack, unpackb as parse_msgpack
except ImportError:
    from umsgpack import packb as to_msgpack, unpackb as parse_msgpack
finally:
    from base64 import b64encode, b64decode

    def dump_term(term, **argd):
        argd.setdefault('use_bin_type', True)
        return b64encode(to_msgpack(to_simple(term), **argd))

    def load_term(encoded, **argd):
        encoding = argd.pop('encoding', 'utf-8')
        term = parse_msgpack(b64decode(encoded), encoding=encoding)
        return parse_term(term, **argd)


try:
    from ujson import dumps as to_json, loads as parse_json
except ImportError:
    try:
        from yajl import Encoder, Decoder
        parse_json = Decoder().decode
        to_json = Encoder().encode
    except ImportError:
        try:
            from jsonlib import write as to_json, read as parse_json
        except ImportError:
            try:
                from cjson import encode as to_json, decode as parse_json
            except ImportError:
                try:
                    from simplejson import dumps as to_json, loads as parse_json  # noqa
                except ImportError:
                    from json import dumps as to_json, loads as parse_json        # noqa


from unicodedata import normalize as _normalize
_punct_regex = regex(r'[\t !"#$%&\'()*\-/<=>?@\[\\\]^_`{|},.:]+')
_file_punct_regex = regex(r'[\t !"#$%&\'*/<=>?@\[\\\]^`{|}:]+')


def normalize_filename(term, mode='NFKD'):
    assert isinstance(term, basestring), 'bad-type: %r' % term
    text = ' '.join(_file_punct_regex.split(path.basename(term)))
    return _normalize(mode, text).encode('ascii', 'ignore')


def normalize(term, mode='NFKD'):
    assert isinstance(term, basestring), 'bad-type: %r' % term
    text = ' '.join(_punct_regex.split(term))
    return _normalize(mode, text).encode('ascii', 'ignore')


def slugify(term, delim='-'):
    return normalize(term).lower().replace(' ', delim)



"""
Toy storage implementation using dataset.
"""

from _term import Hict, parse_term, dump_term as dumps, load_term as loads
from _auth import default_settings, hooks
from _state import get_state
from _admin import Profile
from _store import Record
from _exc import RecordVersionMismatch, DuplicatedRecordID

import _config as conf

from sqlalchemy import LargeBinary, String, Integer, Float as Timestamp
from base64 import b64encode, b64decode
from collections import defaultdict
from time import time as timestamp
from dataset import connect
from os.path import join

# default storage url
MEMORY_DB_URL = 'sqlite:///:memory:'

# default models names
Profile = Profile.__name__
State = '_state_'
Txn = '_txn_'
Prelogin = '_prelogin_'
Abuse = '_abuse_'
Acl = '_acl_'
Settings = '_settings_'


class DSStore(object):

    def __init__(self, uri_prefix=None, **argd):
        self.uri_prefix = uri_prefix
        self.conn_argd = argd
        self.conns = dict()
        self.indexed = set()

    # -- connection & schema ---------------------------------

    def add_model(self, model, add_index):
        if add_index is True:
            self.indexed.add(model)

    def get_connection(self, domain):
        try:
            return self.conns[domain]
        except KeyError:
            if self.uri_prefix is None:
                url = MEMORY_DB_URL
            else:
                url = join(self.uri_prefix, domain)
            conn = connect(url, **self.conn_argd)
            self.init_domain_store_schema(conn)
            return self.conns.setdefault(domain, conn)

    def init_domain_store_schema(self, conn):
        try:
            conn.load_table(Settings)
        except:
            table = conn.create_table(Txn, 'txn', 'String')
            table.create_column('origin', String)
            table.create_column('identity', String)
            table.create_column('tstamp', Timestamp)
            table.create_column('data', LargeBinary)
            table.create_index(['origin', 'tstamp', 'identity'])

            table = conn.create_table(State, 'sid', 'String')
            table.create_column('origin', String)
            table.create_column('identity', String)
            table.create_column('expires', Timestamp)
            table.create_column('data', LargeBinary)
            table.create_index(['origin', 'identity', 'expires'])

            table = conn.create_table(Prelogin, 'sid', 'String')
            table.create_column('origin', String)
            table.create_column('identity', String)
            table.create_column('expires', Timestamp)
            table.create_column('data', LargeBinary)
            table.create_index(['origin', 'identity', 'expires'])

            table = conn.create_table(Abuse)
            table.create_column('origin', String)
            table.create_column('level', String)
            table.create_column('tstamp', Timestamp)
            table.create_column('data', LargeBinary)
            table.create_index(['origin', 'tstamp'])

            table = conn.create_table(Acl)
            table.create_column('type', Integer)
            table.create_column('name', String)
            table.create_column('role', String)
            table.create_column('perm', Integer)
            table.create_index(['type', 'name', 'role'])

            table = conn.create_table(Settings, 'key', 'String')
            table.create_column('value', String)
            table.insert_many([dict(key=k, value=getattr(conf, k.upper()))
                               for k in default_settings], ensure=False)

    def get_model(self, domain, model):
        conn = self.get_connection(domain)
        try:
            table = conn.load_table(model)
        except:
            table = conn.create_table(model, 'uuid', 'String')
            table.create_column('txn', String)
            table.create_column('vsn', Integer)
            table.create_column('data', LargeBinary)
            table.create_index(['txn', 'vsn'])
        else:
            if model in self.indexed:
                missing = [name for name, column
                           in table.table.columns.items()
                           if not column.index and name != 'data']
                if missing:
                    table.create_index(missing)

        return table

    def get_model_hist(self, domain, model):
        conn = self.get_connection(domain)
        model_hist = '%s_hist' % model
        try:
            table = conn.load_table(model_hist)
        except:
            table = conn.create_table(model_hist)
            table.create_column('uuid', String)
            table.create_column('txn', String)
            table.create_column('vsn', Integer)
            table.create_column('data', LargeBinary)
            table.create_index(['uuid', 'txn', 'vsn'])

        return table

    # -- transaction -----------------------------------------

    def begin(self, domain, txn):
        self.get_connection(domain).begin()

    def commit(self, domain, txn, state, log=None):
        self.get_model(domain, '_txn_').insert(dict(
            txn=str(txn), origin=state.origin, tstamp=timestamp(),
            profile=str(state.profile._uuid) if state.profile else None,
            data=dumps(dict(log=log, state=state or get_state()))))
        self.get_connection(domain).commit()

    def rollback(self, domain, txn):
        self.get_connection(domain).rollback()

    # -- crud ------------------------------------------------

    def fetch(self, domain, model, uuid, vsn=None, txn=None):
        table = self.get_model(domain, model)

        if vsn is not None:
            row = table.find_one(uuid=uuid, vsn=vsn)
        else:
            row = table.find_one(uuid=uuid)

        if row is not None:
            return (row['vsn'], row['txn'], loads(row['data']))

    def store(self, domain, txn, inserts=None, updates=None, deletes=None):
        if inserts:
            self.insert(domain, txn, inserts)
        if updates:
            self.update(domain, txn, updates)
        if deletes:
            self.delete(domain, txn, deletes)

    def insert(self, domain, txn, records):
        get_model_hist = self.get_model_hist
        get_model = self.get_model
        rows = defaultdict(list)
        indexed = self.indexed

        for record in records:
            model, uuid = record._model, str(record._uuid)

            if get_model_hist(domain, model).find_one(uuid=uuid):
                raise DuplicatedRecordID(domain, model, uuid)

            row = dict(vsn=record._vsn, uuid=uuid, txn=str(txn),
                       data=dumps(record._dct))

            if model in indexed:
                row.update(record._indexer(record))

            rows[model].append(row)

        for model, rows in rows.iteritems():
            get_model(domain, model).insert_many(rows)

    def update(self, domain, txn, records):
        get_model_hist = self.get_model_hist
        get_model = self.get_model
        indexed = self.indexed

        for record in records:
            model, uuid, vsn = record._model, str(record._uuid), record._vsn
            table = get_model(domain, model)

            stored = table.find_one(uuid=uuid)
            if stored['vsn'] != (vsn - 1):
                raise RecordVersionMismatch(domain, model, uuid, vsn)

            row = dict(vsn=vsn, uuid=uuid, txn=str(txn),
                       data=dumps(record._dct))

            if model in indexed:
                row.update(record._indexer(record))

            get_model_hist(domain, model).insert(stored)
            table.update(row)

    def delete(self, domain, txn, records):
        get_model_hist = self.get_model_hist
        get_model = self.get_model

        for record in records:
            model, uuid, vsn = record._model, str(record._uuid), record._vsn
            table = get_model(domain, model)

            stored = table.find_one(uuid=uuid)
            if stored['vsn'] != -vsn:
                raise RecordVersionMismatch(domain, model, uuid, vsn)

            get_model_hist(domain, model).insert(stored)
            row = dict((c, None) for c in table.columns)
            row.update(uuid=uuid, vsn=vsn, txn=str(txn))
            table.update(row)

    # -- fuzzy & relational query ------------------------------

    def search(self, domain, model, fun):
        raise NotImplementedError

    def select(self, domain, model, fun):
        for row in fun(self.get_model(domain, model)):
            yield Record(model, row.pop('uuid'), vsn=row['vsn'],
                         txn=row['txn'], **loads(row['data']))

    def find(self, domain, model, terms):
        return self.get_model(domain, model).find_one(**terms)

    # -- auth hooks --------------------------------------------

    def _fetch_setting(self, domain, key, fail=None, parser=parse_term):
        table = self.get_model(domain, Settings)
        try:
            row = table.find_one(key=key)
            if parser is not None:
                return parser(row['value'])
            else:
                return row['value']
        except:
            return fail

    def _fetch_record(self, domain, model, **terms):
        row = self.get_model(domain, model).find_one(**terms)
        if row is not None:
            return Record(model, row.pop('uuid'), vsn=row['vsn'],
                          txn=row['txn'], **loads(row['data']))

    def fetch_profile(self, domain, identity, **extra):
        return self._fetch_record(domain, Profile, username=identity, **extra)

    def prelogin_count(self, domain, origin):
        db = self.get_model(domain, Prelogin).database
        sql = '''DELETE FROM %s WHERE origin = :origin
                                  AND expires <= :tstamp'''
        db.query(sql % Prelogin, origin=origin, tstamp=timestamp())
        sql = '''SELECT COUNT(*) AS active FROM %s
                  WHERE origin = :origin'''
        row = db.query(sql % Prelogin, origin=origin).next()
        return row['active']

    def fetch_prelogin(self, origin, domain, sid):
        table = self.get_model(domain, Prelogin)
        db = table.database
        db.begin()
        try:
            enc_sid = b64encode(sid)
            row = table.find_one(sid=enc_sid, origin=origin)
            if row is not None:
                assert sid == b64decode(row.pop('sid'))
                prelogin = Hict(sid=sid, **row)
                prelogin.update(**loads(prelogin.pop('data')))
                return prelogin
        finally:
            table.delete(sid=enc_sid, origin=origin)
            db.commit()

    def store_prelogin(self, origin, domain, sid, identity, expires, **data):
        row = dict(origin=origin, identity=identity, expires=expires,
                   sid=b64encode(sid), data=dumps(data))
        self.get_model(domain, Prelogin).insert(row)

    def load_state(self, domain, sid):
        row = self.get_model(domain, State).find_one(sid=b64encode(sid))
        if row:
            assert sid == b64decode(row.pop('sid'))
            profile = self._fetch_record(domain, Profile, uuid=row.pop('profile'))  # noqa
            state = Hict(sid=sid, **loads(row.pop('data'), dict_type=Hict))
            state.update(profile=profile, **row)
            return state

    def store_state(self, state):
        profile = state.profile
        data = dict(skey=state.skey, ckey=state.ckey, sseed=state.sseed,
                    vkey=state.vkey, pads=state.pads, data=state.data,
                    role=state.role, impersonated_by=state.impersonated_by)
        row = dict(sid=b64encode(state.sid), identity=profile.username,
                   profile=str(profile._uuid), origin=state.origin,
                   expires=state.expires, data=dumps(data))
        self.get_model(state.domain, State).upsert(row, keys=('sid',))

    def destroy_state(self, state):
        self.get_model(state.domain, State).delete(sid=b64encode(state.sid))

    def active_user_sessions(self, domain, identity):
        db = self.get_model(domain, State).database
        sql = '''SELECT COUNT(*) AS active FROM %s
                  WHERE identity = :identity
                    AND expires > :expires'''
        params = dict(identity=identity, expires=timestamp())
        row = db.query(sql % State, **params).next()
        return row['active']

    def destroy_sessions(self, domain, identity):
        self.get_model(domain, State).delete(identity=identity)

    def incr_abuse(self, origin, domain, incr=1, abuse=None):
        data = dumps(abuse) if abuse is not None else abuse
        row = dict(origin=origin, level=incr, tstamp=timestamp(), data=data)
        self.get_model(domain, Abuse).insert(row)

    def abuse_level_by_origin(self, origin, domain):
        watch_period = timestamp() - hooks.abuse_level_watch_period(domain)
        db = self.get_model(domain, Abuse).database
        sql = '''SELECT SUM(level) AS abuse_level FROM %s
                  WHERE origin = :origin
                    AND tstamp >= :watch_period'''
        params = dict(origin=origin, watch_period=watch_period)
        row = db.query(sql % Abuse, **params).next()
        return row['abuse_level'] or 0

    def session_timeout(self, domain):
        return self._fetch_setting(domain, 'session_timeout')

    def max_prelogin_count(self, domain):
        return self._fetch_setting(domain, 'max_prelogin_count')

    def abuse_level_watch_period(self, domain):
        return self._fetch_setting(domain, 'abuse_level_watch_period')

    def max_origin_abuse_level(self, domain):
        return self._fetch_setting(domain, 'max_origin_abuse_level')

    def max_active_user_sessions(self, domain, identity):
        return self._fetch_setting(domain, 'max_active_user_sessions')

    def has_permission(self, state, call):
        return True



"""
Handle sessions in execution stack.
"""
__all__ = ['get_state']

from time import time as timestamp
from inspect import currentframe

from _auth import hooks as auth
from _term import Hict
import _exc as exc


class State(Hict):
    pass


def new_state(domain='localhost', **data):
    default = Hict(origin='127.0.0.1', domain=domain, service=None,
                   expires=timestamp() + auth.session_timeout(domain),
                   impersonated_by=None, profile=None,
                   sid=auth.gen_session_id())
    default.update(data)
    return State(default)


def get_state(**data):
    frame = caller = currentframe().f_back
    while frame:
        _state_ = frame.f_locals.get('_state_', None)
        if not _state_:
            frame = frame.f_back
            continue
        elif caller is not frame:
            caller.f_locals['_state_'] = _state_
        break
    else:
        _state_ = new_state(**data)
        caller.f_locals['_state_'] = _state_
    return _state_


def load_state(sid, **data):
    _state_ = auth.load_state(sid)
    if _state_:
        if _state_.expires > timestamp():
            _state_.profile = auth.fetch_profile(_state_.profile)
        else:
            raise exc.SessionTimeout
    else:
        _state_ = new_state()
        store_state(_state_)
    return _state_


def store_state(state):
    state.expires = timestamp() + auth.session_timeout(state.domain)
    state = State(state)
    state.profile = getattr(state.profile, '_uuid', None)
    auth.store_state(state)


def clear_state(state):
    state.data.clear()
    state.expires = 0



"""
Storage proxy machinery. Not an ORM.
"""
__all__ = ['add_storage', 'add_model', 'Record', 'DataStore', 'storage']

from collections import defaultdict
from _term import parse_term, gen_uuid, is_class
from _auth import hooks, default_settings
from _state import get_state
import _exc as exc

from time import time as timestamp

import logging
logger = logging.getLogger('fanery.storage')

storage_reg = dict()
model_reg = dict()


class DataStore():

    def __init__(self, storage, fuzzy=None, relational=None, permission=None,
                 state=None, abuse=None, profile=None, settings=None):
        """
        When building the storage strategy for your models keep in mind:

        1. It should offer CRUD + relational & fuzzy search.
        2. Indexing and persistency should be handled by separate and
           specialized database engines.
        3. Each model/data-type may be handled by a different database engine.
        4. Avoid engines that only offer eventual consistency.
        5. Avoid engines that require stop of operation during backup.
        6. Avoid propietary/closed-source Software solutions.
        7. Use real elastic database and search engines.
        """
        relay = self.__dict__

        # attach data persistence behaviours

        relay.update(store=storage.store,
                     fetch=storage.fetch,
                     begin=storage.begin,
                     commit=storage.commit,
                     rollback=storage.rollback)

        # attach data indexing behaviours

        indexers = set()

        if fuzzy is None:
            fuzzy = storage
        elif fuzzy is not storage:
            indexers.add(fuzzy)

        if relational is None:
            relational = storage
        elif relational is not storage:
            indexers.add(relational)

        indexers = tuple(indexers)

        relay.update(search=fuzzy.search,
                     find=relational.find,
                     select=relational.select,
                     indexers=indexers)

        # attach schema creation delegator

        add_index = bool(relational is storage or fuzzy is storage)

        def add_model(model):
            storage.add_model(model, add_index=add_index)
            for indexer in indexers:
                indexer.add_model(model)

        relay.update(add_model=add_model)

        # register auth hooks

        if permission:
            hooks.has_permission = permission.has_permission

        if settings:
            hooks.update((name, getattr(settings, name))
                         for name in default_settings
                         if hasattr(settings, name))

        if profile:
            hooks.fetch_profile = profile.fetch_profile

        if abuse:
            hooks.update(incr_abuse=abuse.incr_abuse,
                         abuse_level_by_origin=abuse.abuse_level_by_origin)

        if state:
            hooks.update(load_state=state.load_state,
                         store_state=state.store_state,
                         destroy_state=state.destroy_state,
                         active_user_sessions=state.active_user_sessions,
                         destroy_user_sessions=state.destroy_sessions,
                         fetch_prelogin=state.fetch_prelogin,
                         store_prelogin=state.store_prelogin,
                         prelogin_count=state.prelogin_count)


def add_storage(storage, *models):
    for model in (models or model_reg.keys()):
        model = getattr(model, '__name__', model)
        assert model in model_reg, ('unknown-model', model)
        storage_reg[model] = storage
        storage.add_model(model)


def add_model(model, validator=None, initializer=None, indexer=None):
    name = getattr(model, '__name__', model)
    model_reg[name] = (getattr(model, 'validate', validator),
                       getattr(model, 'initialize', initializer),
                       getattr(model, 'index', indexer))


def get_storage(model):
    assert model in storage_reg, model
    return storage_reg[model]


def get_validator(model):
    return model_reg[model][0] if model in model_reg else None


def get_initializer(model):
    return model_reg[model][1] if model in model_reg else None


def get_indexer(model):
    return model_reg[model][2] if model in model_reg else None


def get_fields(model):
    return model_reg[model][3] if model in model_reg else None


class Aict(dict):
    """
    Hierarchical dotted dictionary with value auto-parsing.
    """

    def __missing__(self, key):
        term = self[key] = Aict()
        return term

    def __setitem__(self, key, value):
        dict.__setitem__(self, key, parse_term(value, dict_type=Aict))

    __setattr__ = __setitem__
    __getattr__ = dict.__getitem__
    __delattr__ = dict.__delitem__


# record marks
UNDEF, INSERT, UPDATE, DELETE = range(4)


class Record(object):

    def __init__(self, model, uuid, vsn, txn=None, _dct_=None, **dct):
        model = getattr(model, '__name__', model)
        self._store = store = get_storage(model)
        if _dct_:
            self._dct = parse_term(_dct_, dict_type=Aict)
        elif dct:
            self._dct = parse_term(dct, dict_type=Aict)
        elif vsn > 0:
            record_data = store.get(uuid, txn, vsn, txn)
            if not record_data:
                raise exc.RecordNotFound(model, uuid, vsn)
            txn, data = record_data
            self._dct = parse_term(data, dict_type=Aict)
        self._mark = UNDEF
        self._model = model
        self._uuid = parse_term(uuid)
        self._vsn = parse_term(vsn)
        self._txn = parse_term(txn)
        self._indexer = get_indexer(model)
        initialize = get_initializer(model)
        if initialize:
            initialize(self)

    def __str__(self):
        return '<Record "%s" %s:%d>' % (self._model, self._uuid, self._vsn)
    __repr__ = __str__

    def __getattr__(self, key):
        return self._dct[key]

    def __setattr__(self, key, value):
        dct = self.__dict__ if key[0] == '_' else self._dct
        dct[key] = parse_term(value, dict_type=Aict)

    def __delattr__(self, key):
        if key[0] != '_':
            del self._dct[key]

    def update(self, *args, **argd):
        self._dct.update(*args if args else argd)
        return self

    def merge(self, *args, **argd):
        dct, items = self._dct, args or argd.iteritems()
        dct.update((k, v) for k, v in items if k not in dct)
        return self


class storage(object):
    """
    The idea here is to handle all storage activities as single
    transactional unit of works inside a with statement.

        with storage("some description") as db:
            do something with db
    """

    def __init__(self, log_message=None):
        self._start = timestamp()
        self._state = state = get_state()
        self._domain = state.domain
        self._log = log_message
        self._txn = gen_uuid()
        self._cache = dict()
        self._dirty = defaultdict(set)

        logger.debug('%s %s' % (self._txn, state.service or 'init'))

    def __enter__(self):
        logger.debug("%s enter %s" % (self._txn, self._log or ''))

        return self

    def __exit__(self, type, value, traceback):
        try:
            if self._cache:
                if isinstance(value, Exception):
                    self.__rollback()
                else:
                    self.__commit()
        finally:
            logger.debug("%s %0.6f exit" % (self._txn, timestamp() - self._start))    # noqa

    def __commit(self):
        domain = self._domain
        txn = self._txn

        dirty = self._dirty
        dbms = dirty.keys()

        _start = timestamp()
        try:
            for dbm in dbms:
                dbm.begin(domain, txn)
        finally:
            logger.debug("%s %0.6f begin" % (txn, timestamp() - _start))

        try:
            for dbm, records in dirty.iteritems():
                inserts = set()
                updates = set()
                deletes = set()

                for record in records:
                    mark = record._mark
                    _vsn, _txn = record._vsn, record._txn
                    if mark is INSERT:
                        record._vsn = 1
                        inserts.add(record)
                    elif mark is UPDATE:
                        record._vsn += 1
                        updates.add(record)
                    elif mark is DELETE:
                        record._vsn *= -1
                        deletes.add(record)
                    record._mark = UNDEF
                    record._txn = txn

                _start = timestamp()
                try:
                    dbm.store(domain, txn,
                              inserts=inserts,
                              updates=updates,
                              deletes=deletes)
                finally:
                    logger.debug("%s %0.6f store ins=%d upd=%d del=%d" % (
                        txn, timestamp() - _start, len(inserts), len(updates), len(deletes)))    # noqa

                _start = timestamp()
                try:
                    for indexer in dbm.indexers:
                        indexer.index(domain, txn,
                                      inserts=inserts,
                                      updates=updates,
                                      deletes=deletes)
                finally:
                    logger.debug("%s %0.6f index" % (txn, timestamp() - _start))  # noqa

        except Exception:
            self.__rollback()
            raise

        else:
            _start = timestamp()
            try:
                for dbm in dbms:
                    dbm.commit(domain, txn, self._state, self._log)
            finally:
                logger.debug("%s %0.6f commit" % (txn, timestamp() - _start))

    def __rollback(self):
        _start = timestamp()
        try:
            for dbm in self._dirty.keys():
                dbm.rollback(self._domain, self._txn)
        finally:
            logger.debug("%s rollback [%0.6f]" % (self._txn, timestamp() - _start)) # noqa

    def fetch(self, model, uuid, vsn=None, txn=None, acl=None, fail=None):
        logger.debug("%s fetch %s %s" % (self._txn, model, uuid))

        model = getattr(model, '__name__', model)
        domain = self._domain
        cache = self._cache

        record = cache.get((model, uuid, vsn), None)
        if record and record._mark is DELETE:
            record = None
        elif record is None:
            storage = get_storage(model)
            record = storage.fetch(domain, model, uuid, vsn, txn)
            if record and not isinstance(Record, record):
                record = Record(model, uuid, *record)

        if acl and record and acl(record) is not True:
            record = None

        if record is not None:
            return cache.setdefault((model, uuid, record._vsn), record)
        elif issubclass(fail, Exception):
            raise fail
        elif callable(fail):
            return fail()
        else:
            return fail

    def query(self, method, model, fun, acl=None, raw=False, load=parse_term):
        _start = timestamp()
        try:
            model = getattr(model, '__name__', model)
            index = get_storage(model)
            domain = self._domain
            cache = self._cache

            for term in getattr(index, method)(domain, model, fun):
                if raw is True:
                    yield load(term) if load else term

                if isinstance(term, Record):
                    uuid, vsn, score = term._uuid, term._vsn, None
                    record = cache.setdefault((domain, model, uuid, vsn), term)
                else:
                    size = len(term)
                    if size == 4:
                        uuid, vsn, txn, score = term
                    elif size == 3:
                        uuid, vsn, txn = term
                        score = None
                    else:
                        uuid, vsn = term
                        txn = score = None
                    record = cache.get((domain, model, uuid, vsn), None)

                if record is None:
                    record = self.fetch(model, uuid, vsn, txn, acl)
                    if record is None:
                        continue
                    cache[(domain, model, uuid, vsn)] = record
                elif record._mark is DELETE or (acl and acl(record) is not True):   # noqa
                    continue
                yield (score, record) if score is not None else record
        finally:
            logger.debug("%s %0.6f %s %s %s" % (self._txn, timestamp() - _start, method, model, fun))   # noqa

    def search(self, model, fun, acl=None):
        """Fuzzy search."""
        return self.query('search', model, fun, acl)

    def select(self, model, fun, acl=None):
        """Relational query."""
        return self.query('select', model, fun, acl)

    def select_one(self, model, fun, acl=None, fail=None):
        for idx, record in enumerate(self.select(model, fun, acl)):
            if idx > 0:
                raise exc.MultipleRecordsFound(model)

        try:
            return record
        except:
            if is_class(fail) and issubclass(fail, Exception):
                raise fail
            elif callable(fail):
                return fail()
            else:
                return fail

    def insert(self, model, *args, **argd):
        _start = timestamp()
        try:
            model = getattr(model, '__name__', model)
            uuid, vsn = gen_uuid(), 0

            record = Record(model, uuid, vsn, self._txn, *args, **argd)
            self.validate(record)
            record._mark = INSERT
            self._dirty[record._store].add(record)
            return self._cache.setdefault((model, uuid, vsn), record)
        finally:
            logger.debug("%s %0.6f insert %s" % (self._txn, timestamp() - _start, model))   # noqa

    def update(self, *records):
        _start = timestamp()
        try:
            for record in records:
                self.validate(record)
                mark = record._mark
                if mark is DELETE:
                    raise exc.RecordMarkedForDeletion(record)
                elif record._vsn < 0:
                    raise exc.DeletedRecord(record)
                elif mark is INSERT:
                    pass
                else:
                    record._mark = UPDATE
                    self._dirty[record._store].add(record)
        finally:
            logger.debug("%s %0.6f update %d" % (self._txn, timestamp() - _start, len(records)))    # noqa

    def delete(self, *records):
        _start = timestamp()
        try:
            for record in records:
                if record._mark is INSERT:
                    raise exc.RecordMarkedForInsertion(record)
                elif record._vsn < 0:
                    raise exc.DeletedRecord(record)
                elif record._vsn == 0:
                    raise exc.UnstoredRecord(record)
                record._mark = DELETE
                self._dirty[record._store].add(record)
        finally:
            logger.debug("%s %0.6f delete %d" % (self._txn, timestamp() - _start, len(records)))    # noqa

    def validate(self, record):
        _start = timestamp()
        try:
            validator = get_validator(record._model)
            if validator:
                errors = validator(record)
                if errors:
                    raise exc.RecordValidationError(record, errors)
        finally:
            logger.debug("%s %0.6f validate %s" % (self._txn, timestamp() - _start, record._model))    # noqa



"""
Third party profiling decorators.
"""
__all__ = ['timecall', 'coverage', 'memory_profile', 'line_profile']

from functools import wraps

try:
    from memory_profiler import profile as memory_profile
except ImportError:
    pass

try:
    from line_profiler import LineProfiler

    def line_profile(func):
        p = LineProfiler()
        f = p(func)

        def wrapper(*args, **argd):
            try:
                return f(*args, **argd)
            finally:
                p.print_stats()
        return wrapper
except ImportError:
    pass

try:
    from profilehooks import profile as _profile

    def profile(f, immediate=True):
        return _profile(f, immediate=immediate)

except ImportError:
    raise
    try:
        from pstats import Stats

        try:
            from cProfile import Profile
        except ImportError:
            from profile import Profile

        def profile(f, immediate=True):
            @wraps(f)
            def wrapper(*args, **argd):
                profiler = Profile()
                try:
                    profiler.enable()
                    return f(*args, **argd)
                finally:
                    profiler.disable()
                    stats = Stats(profiler)
                    stats.strip_dirs()
                    stats.sort_stats('cumulative')
                    stats.print_stats(f.__name__)
            return wrapper
    except ImportError:
        pass

try:
    from profilehooks import timecall
except ImportError:
    from time import time

    def timecall(f, immediate=True):
        @wraps(f)
        def wrapper(*args, **argd):
            try:
                start = time()
                return f(*args, **argd)
            finally:
                print "\n  %s (%s:%d):\n    %.3f seconds\n" % (
                    f.func_name, f.func_code.co_filename,
                    f.func_code.co_firstlineno,
                    time() - start)
        return wrapper

try:
    from profilehooks import coverage
except ImportError:
    pass



import logging
logging.basicConfig(level=logging.ERROR)

from fanery import (
    PyClient, WsgiClient,
    service,
    get_state,
    rand_num, rand_bytes,
    Hict, is_date,
    exc
)


# -- services definition -----------------------------


@service(safe=False, ssl=False)
def echo(term, *terms):
    return term if not terms else (term,) + terms


@service(ssl=False, auto_parse=False)
def profile():
    profile = get_state().profile
    return profile._indexer(profile)


def _outer_state(state):

    def _outer_inner_state():
        assert state is get_state()

    _outer_inner_state()
    return get_state()


@service(auto_parse=False)
def check_state():
    from fanery._state import get_state

    state = get_state()

    def _inner_state():

        def _inner_inner_state():
            assert get_state() is state

        _inner_inner_state()
        return get_state()

    assert state is _outer_state(state)
    assert state is _inner_state()
    assert state is get_state()

    assert state.service is check_state.urlpath


# -- test helpers -----------------------------------


def setup(db):
    from fanery import admin, get_state
    from fanery._auth import hooks

    state = get_state()

    admin.setup(db)
    admin.init('admin', 'admin', domain=state.domain)
    profile = hooks.fetch_profile(state.domain, 'admin')
    assert profile

    hooks.destroy_user_sessions(state.domain, 'admin')
    try:
        hooks.reset_origin_abuse_level(state.origin, state.domain)
    except:
        pass


def _login(client):
    assert client.login('admin', 'admin') is True


def _echo(client, terms=10):
    r = dict((rand_bytes(8), rand_num(dec=False)) for _ in xrange(terms))
    for k, v in r.items():
        for t in (k.encode('hex'), v, [v, v / 2]):
            ret = client.call('echo', term=v)
            assert ret == v, (ret, v)


def _profile(client):
    _login(client)
    data = Hict(client.safe_call('profile'))
    assert (data.username == 'admin' and
            data.domain == 'localhost' and
            data.email == 'admin@example.com' and
            'password_hash' not in data and
            'password_salt' not in data), data


def _check_state(client):
    _login(client)
    client.safe_call('check_state')


def _logout(client):
    _login(client)
    client.safe_call('check_state')
    assert client.logout() is True

    from fanery._auth import hooks
    abuse_level = hooks.abuse_level_by_origin

    state = get_state()
    origin = state.origin
    domain = state.domain

    level = abuse_level(origin, domain)
    try:
        client.safe_call('check_state')
        raise Exception('should raise UnknownSession')
    except exc.UnknownSession:
        assert level < abuse_level(origin, domain)


# -- test functions ------------------------------


def test_urlpath():
    assert echo.urlpath == 'echo'
    assert profile.urlpath == 'profile'


def test_safe_param():
    assert echo.safe is False
    assert profile.safe is True


def test_ssl_param():
    assert echo.ssl is False
    assert profile.ssl is False


def test_auto_parse_param():
    assert echo.auto_parse is True
    assert profile.auto_parse is False


def test_auto_parse_on_call():
    assert echo('1') == 1
    assert echo('1.1') == 1.1
    assert is_date(echo('1990-01-01'))


def test_lookup():
    from fanery._term import to_json, to_str
    from fanery._service import lookup, static, registry

    assert lookup('prefix/echo/1/2/3.json') == (echo, ['1', '2', '3'], '.json', to_json)  # noqa
    assert lookup('/echo.doc') == (echo, [], '.doc', False)
    assert lookup('/a/profile') == (profile, [], '', None)
    assert lookup('profile.txt') == (profile, [], '.txt', to_str)

    if '' not in registry:
        assert lookup('/index.html') == (None, ['index.html'], '.html', None)
        serve_file = static('/', '.')
        assert lookup('/index.html') == (serve_file, ['index.html'], '.html', False)    # noqa


def test_consume():
    from fanery._service import consume

    _state_ = get_state()
    assert consume(_state_, 'prefix/echo/1/2/3') == (echo, '', (1, 2, 3))
    assert consume(_state_, 'prefix/echo/1/2/3.json') == (echo, '.json', '[1,2,3]')     # noqa


def test_py_login():
    _login(PyClient())


def test_wsgi_login():
    _login(WsgiClient())


def test_py_echo():
    _echo(PyClient())


def test_wsgi_echo():
    _echo(WsgiClient())


def test_py_check_state():
    try:
        _check_state(PyClient())
        raise Exception('should raise RequireSSL')
    except exc.RequireSSL:
        _check_state(PyClient(ssl=True))


def test_wsgi_check_state():
    try:
        _check_state(WsgiClient())
        raise Exception('should raise RequireSSL')
    except exc.RequireSSL:
        _check_state(WsgiClient(ssl=True))


def test_py_profile():
    _profile(PyClient())


def test_wsgi_profile():
    _profile(WsgiClient())


def test_py_logout():
    _logout(PyClient(ssl=True))


def test_wsgi_logout():
    _logout(WsgiClient(ssl=True))


# -- tests run -------------------------------------------


def main():
    from fanery import DataStore, dbm, timecall  # , memory_profile, line_profile   # noqa
    from random import shuffle

    tests = [(k, v) for k, v in globals().iteritems()
             if k.startswith('test_') and callable(v)]

    for DS in (dbm.MemDictStore,
               #dbm.DSStore,
               #lambda: dbm.RTStore('192.168.122.61', indexer='self'),
               #lambda: dbm.HDStore('192.168.122.91', 2001, indexer='self'),
               ):

        ds = DS()
        storage = DataStore(storage=ds,
                            permission=ds,
                            state=ds,
                            abuse=ds,
                            profile=ds,
                            settings=ds)

        shuffle(tests)
        timecall(setup)(storage)
        for name, test in tests:
            #memory_profile(test)()
            #line_profile(test)()
            timecall(test)()

if __name__ == '__main__':
    main()



"""
Fanery wsgi handler.
"""
__all__ = ['build_handler']

from mimetypes import guess_type, add_type
add_type('application/json', '.json')

from webob import Response
from webob.dec import wsgify
from webob.static import FileApp, DirectoryApp
from webob.exc import HTTPBadRequest, HTTPNotFound, HTTPNotImplemented

from _term import (
    Hict, normalize_filename, parse_json, to_simple, to_json,
    is_file_path, is_dir_path, is_sequence, is_generator
)
from _service import consume
from _state import State
from _exc import NotFound
from _version import __version__

import _config as conf

SERVER = 'Fanery/%s' % __version__
CONTENT_TYPE = 'text/plain'
CHARSET = 'utf8'
ENCODING_ERROR = 'Wrong request encoding, must be %s.' % CHARSET
STRICT_HTML_HEADERS = {
    'X-Frame-Options': 'SAMEORIGIN',
    'X-XSS-Protection': '1; mode=block',
    'Content-Security-Policy': "default-src 'self' 'unsafe-eval'",
}

from tempfile import NamedTemporaryFile
from os import link, unlink, chmod
from os.path import basename, join
from stat import S_IRUSR
from cgi import FieldStorage

import logging
logger = logging.getLogger('fanery.wsgi')


class BrokenUpload:
    pass


def _cgi_FieldStorage__make_file__patch(self, binary=None):
    '''monkey patch for FieldStorage.make_file

    Use named temporary file which allow then to create
    a permanent link after successfull upload and so
    avoid copy overhead.'''
    return NamedTemporaryFile(dir=conf.UPLOAD_DIRPATH)

FieldStorage.make_file = _cgi_FieldStorage__make_file__patch


def _permanent(temp_file):
    '''Create read-only hard link and close the given temporary file.'''
    dest_path = join(conf.UPLOAD_DIRPATH, basename(temp_file.filename))
    link(temp_file.file.name, dest_path)
    chmod(dest_path, S_IRUSR)
    return dest_path


def _param_value(term, links):
    if not isinstance(term, FieldStorage):
        return term
    elif term.done == -1:
        return BrokenUpload
    else:
        link = _permanent(term)
        links.add(link)
        return link


@wsgify
def handler(req):

    # parse params and consume service

    try:
        req.charset = CHARSET
        argd, params, path, links = Hict(), req.params, req.path, set()

        for key in set(params.keys()):
            if not key.startswith('_'):
                if key.endswith('[]'):
                    values = (_param_value(term, links)
                              for term in params.getall(key))
                    argd[key[:-2]] = tuple(value for value in values
                                           if value is not BrokenUpload)
                else:
                    value = _param_value(params[key], links)
                    if value is not BrokenUpload:
                        argd[key] = value

        if '_json_' in params:
            argd.update(parse_json(params['_json_']))

        _state_ = State(domain=req.host.split(':', 1)[0],
                        origin=req.client_addr or '127.0.0.1',
                        sid=None, profile=None, role=None,
                        ssl=bool(req.headers.get('X-Forwarded-Proto', req.scheme) == 'https'))                  # noqa

        fun, ext, ret = consume(_state_, path, **argd)
        links.clear()   # keep uploaded files
    except NotFound, e:
        return HTTPNotFound(server=SERVER)
    except NotImplementedError, e:
        return HTTPNotImplemented(server=SERVER)
    except UnicodeDecodeError:
        return HTTPBadRequest(ENCODING_ERROR, server=SERVER)
    except Exception, e:
        res = HTTPBadRequest(content_type='text/plain', server=SERVER)
        error = dict(exc=e.__class__.__name__, err=e.args)
        res.body = to_json(to_simple(error))
        return res
    finally:
        # cleanup uploaded files on error
        for link in links:
            unlink(link)

    # build answer

    headers = {'Content-Security-Policy': "default-src 'none'"}

    if fun.ssl is True:
        headers['Strict-Transport-Security'] = "max-age=31536000; includeSubDomains"                        # noqa

    if fun.static is True:

        if isinstance(ret, file):
            if ret.closed is False and fun.auto_close_file is True:
                ret.close()
            ret = ret.name

        headers['X-Content-Type-Options'] = 'nosniff'

        if fun.cache is not True:
            headers['Cache-Control'] = 'max-age=0, must-revalidate, no-cache, no-store'                     # noqa
            headers['Pragma'] = 'no-cache'
            headers['Expires'] = '0'

        if fun.accel is not None:
            headers['X-Accel-Redirect' if fun.accel == 'nginx' else 'X-Sendfile'] = ret                     # noqa
            return Response(server=SERVER, headerlist=headers.items())

        elif is_file_path(ret):

            if fun.force_download is True:
                headers['Content-Type'] = 'application/octet-stream'

                # Content-Disposition security considerations
                # http://tools.ietf.org/html/rfc6266#section-7
                if fun.content_disposition is True:
                    filename = normalize_filename(ret)
                    headers['Content-Disposition'] = 'attachment; filename="%s"' % filename                 # noqa
            else:
                headers['Content-Type'] = content_type = guess_type(ret)[0] or CONTENT_TYPE                 # noqa

                if content_type == 'text/html':
                    headers.update(STRICT_HTML_HEADERS)

            return FileApp(ret, server=SERVER, headerlist=headers.items())

        elif is_dir_path(ret):
            return DirectoryApp(ret, server=SERVER)

        else:
            return HTTPNotFound(server=SERVER)

    else:
        content_type = guess_type(path)[0] or CONTENT_TYPE

        if content_type == 'text/html':
            headers.update(STRICT_HTML_HEADERS)

        res = Response(server=SERVER, charset=CHARSET,
                       headerlist=headers.items(),
                       content_type=content_type)

        if isinstance(ret, str):
            res.body = ret
        elif isinstance(ret, unicode):
            res.unicode_body = ret
        elif is_sequence(ret) or is_generator(ret):
            res.app_iter = ret
        elif isinstance(ret, file):
            res.body_file = ret
        else:
            res.body = repr(ret)

        cache = fun.cache
        if not cache:
            res.cache_expires(0)
        elif isinstance(cache, int):
            res.cache_expires(cache)
        else:
            res.cache_expires(86400)

        return res


def build_handler(wsgi_app=handler, profile=False, **argd):

    if profile is True:
        from linesman import middleware
        open(middleware.ENABLED_FLAG_FILE, 'w').close()
        wsgi_app = middleware.make_linesman_middleware(wsgi_app, **argd)
        logger.info("wsgi profiler path is %s" % wsgi_app.profiler_path)

    return wsgi_app



from _mddb import MemDictStore  # noqa

try:
    from _dsdb import DSStore   # noqa
except ImportError:
    pass

try:
    from _hddb import HDStore   # noqa
except ImportError:
    pass

try:
    from _rtdb import RTStore   # noqa
except ImportError:
    pass



"""
Convenient functions for random terms generation.
"""
__all__ = ['rand_bytes', 'rand_uuid', 'rand_num', 'rand_int',
           'random', 'rand_num_range']

from _crypto import nacl_random as rand_bytes

from random import randint as rand_int, random, randrange as rand_range

from uuid import uuid4 as rand_uuid


def rand_num(size=4, sign=True, dec=True):
    i = int(rand_bytes(rand_int(1, size)).encode('hex'), 16)
    m = -1 if int(rand_bytes(1).encode('hex'), 16) % 2 else 1
    return m * i * ((1.0 + random()) if dec else 1)


def rand_num_range(size=4, step=1, sign=True):
    a, b = rand_num(size, dec=False), rand_num(size, dec=False)
    return rand_range(a, b, step=step)



"""
Toy storage implementation using rethinkdb.
"""

from _term import Hict, parse_term, dump_term as dumps, load_term as loads
from _auth import default_settings, hooks
from _state import get_state
from _admin import Profile
from _store import Record
from _exc import RecordVersionMismatch, DuplicatedRecordID

import _config as conf

from base64 import b64encode, b64decode
from collections import defaultdict
from time import time as timestamp
import rethinkdb as rtdb

# default models names
Profile = Profile.__name__
State = '_state_'
Txn = '_txn_'
Prelogin = '_prelogin_'
Abuse = '_abuse_'
Acl = '_acl_'
Settings = '_settings_'


class RTModel(object):
    pass


class RTStore(object):

    transactional = False

    def __init__(self, host='localhost', port=28015, indexer=None):
        self.indexer = self if indexer == 'self' else indexer
        self.conn = rtdb.connect(host=host, port=port)

        # setup hooks
        hooks.update(load_state=self.load_state,
                     store_state=self.store_state,
                     destroy_state=self.destroy_state,
                     session_timeout=self.session_timeout,
                     active_user_sessions=self.active_user_sessions,
                     max_active_user_sessions=self.max_active_user_sessions,
                     destroy_user_sessions=self.destroy_sessions,
                     fetch_profile=self.fetch_profile,
                     has_permission=self.has_permission,
                     fetch_prelogin=self.fetch_prelogin,
                     store_prelogin=self.store_prelogin,
                     incr_abuse=self.incr_abuse,
                     abuse_level_by_origin=self.abuse_level_by_origin,
                     max_origin_abuse_level=self.max_origin_abuse_level,
                     abuse_level_watch_period=self.abuse_level_watch_period,
                     prelogin_count=self.prelogin_count)

    # -- connection & schema ---------------------------------

    def add_model(self, model):
        pass

    def get_connection(self, domain):
        conn = self.conn
        try:
            rtdb.db_create(domain).run(conn)
            self.init_domain_store_schema(conn, domain)
        except:
            pass
        return conn

    def init_domain_store_schema(self, conn, domain):
        db = rtdb.db(domain)

        tables = set(db.table_list().run(conn))

        for table, (pk, indexes) in {
                Txn: ('txn', ['origin', 'tstamp', 'identity']),
                State: ('sid', ['origin', 'identity', 'expires']),
                Prelogin: ('sid', ['origin', 'identity', 'expires']),
                Abuse: (None, ['origin', 'tstamp']),
                Acl: (None, ['type', 'name', 'role']),
                Settings: ('key', [])}.items():

            if table not in tables:
                if pk is not None:
                    db.table_create(table, primary_key=pk).run(conn)
                else:
                    db.table_create(table).run(conn)

                for index in indexes:
                    db.table(table).index_create(index).run(conn)

                if table == Settings:
                    db.table(Settings).insert([
                        dict(key=k, value=getattr(conf, k.upper()))
                        for k in default_settings]).run(conn)

    def get_model(self, domain, model):
        # db name can only have alphanum and underscores
        domain = domain.encode('hex')

        conn = self.get_connection(domain)
        db = rtdb.db(domain)

        try:
            db.table_create(model, primary_key='uuid').run(conn)
            db.table(model).index_create('txn').run(conn)
            db.table(model).index_create('vsn').run(conn)
        except:
            pass

        return conn, db.table(model)

    def get_model_hist(self, domain, model):
        # db name can only have alphanum and underscores
        domain = domain.encode('hex')

        conn = self.get_connection(domain)
        model_hist = '%s_hist' % model
        db = rtdb.db(domain)

        try:
            db.table_create(model_hist).run(conn)
            db.table(model_hist).index_create('uuid').run(conn)
            db.table(model_hist).index_create('txn').run(conn)
            db.table(model_hist).index_create('vsn').run(conn)
        except:
            pass

        return conn, db.table(model_hist)

    # -- transaction -----------------------------------------

    def begin(self, domain, txn):
        pass

    def commit(self, domain, txn, state, log=None):
        conn, table = self.get_model(domain, Txn)
        profile = str(state.profile._uuid) if state.profile else None
        data = dumps(dict(log=log, state=state or get_state()))
        table.insert(dict(txn=str(txn), origin=state.origin,
                          tstamp=timestamp(), profile=profile,
                          data=data)).run(conn)

    def rollback(self, domain, txn):
        pass

    # -- crud ------------------------------------------------

    def fetch(self, domain, model, uuid, vsn=None, txn=None):
        conn, table = self.get_model(domain, model)

        if vsn is not None:
            row = table.find_one(uuid=uuid, vsn=vsn)
        else:
            row = table.find_one(uuid=uuid)

        if row is not None:
            return (row['vsn'], row['txn'], loads(row['data']))

    def store(self, domain, txn, insert=None, update=None, delete=None):
        if insert:
            self.insert(domain, txn, insert)
        if update:
            self.update(domain, txn, update)
        if delete:
            self.delete(domain, txn, delete)

    def insert(self, domain, txn, records):
        index = bool(self.indexer is self)
        get_model_hist = self.get_model_hist
        get_model = self.get_model
        rows = defaultdict(list)
        tables = dict()

        for record in records:
            model, uuid = record._model, str(record._uuid)

            if model not in tables:
                conn, table = tables[model] = get_model_hist(domain, model)
            else:
                conn, table = tables[model]

            if table.get(uuid).run(conn) is not None:
                raise DuplicatedRecordID(domain, model, uuid)

            row = dict(vsn=record._vsn, uuid=uuid, txn=str(txn),
                       data=dumps(record._dct))

            if index is True:
                row.update(record._indexer(record))

            rows[model].append(row)

        for model, rows in rows.iteritems():
            conn, table = get_model(domain, model)
            table.insert(rows).run(conn)

    def update(self, domain, txn, records):
        index = bool(self.indexer is self)
        get_model_hist = self.get_model_hist
        get_model = self.get_model
        tables = dict()

        for record in records:
            model, uuid, vsn = record._model, str(record._uuid), record._vsn

            if model not in tables:
                conn, table = get_model(domain, model)
                _, hist = get_model_hist(domain, model)
                tables[model] = (conn, table, hist)
            else:
                conn, table, hist = tables[model]

            stored = table.find_one(uuid=uuid).run(conn)
            if stored['vsn'] != (vsn - 1):
                raise RecordVersionMismatch(domain, model, uuid, vsn)

            row = dict(vsn=vsn, uuid=uuid, txn=str(txn),
                       data=dumps(record._dct))

            if index is True:
                row.update(record._indexer(record))

            hist.insert(stored).run(conn)
            table.update(row).run(conn)

    def delete(self, domain, txn, records):
        get_model_hist = self.get_model_hist
        get_model = self.get_model
        tables = dict()

        for record in records:
            model, uuid, vsn = record._model, str(record._uuid), record._vsn

            if model not in tables:
                conn, table = get_model(domain, model)
                _, hist = get_model_hist(domain, model)
                tables[model] = (conn, table, hist)
            else:
                conn, table, hist = tables[model]

            stored = table.find_one(uuid=uuid).run(conn)
            if stored['vsn'] != -vsn:
                raise RecordVersionMismatch(domain, model, uuid, vsn)

            hist.insert(stored).run(conn)
            row = dict(uuid=uuid, vsn=vsn, txn=str(txn))
            table.update(row).run(conn)

    # -- fuzzy & relational query ------------------------------

    def search(self, domain, model, fun):
        if (self.indexer or self) is self:
            raise NotImplementedError
        else:
            yield self.indexer.search(domain, model, fun)

    def select(self, domain, model, fun):
        if (self.indexer or self) is self:
            conn, table = self.get_model(domain, model)
            table.find = lambda **qry: table.filter(qry)
            for row in fun(table).run(conn):
                yield Record(model, row.pop('uuid'), vsn=row['vsn'],
                             txn=row['txn'], **loads(row['data']))
        else:
            yield self.indexer.select(domain, model, fun)

    # -- auth hooks --------------------------------------------

    def _fetch_setting(self, domain, key, fail=None, parser=parse_term):
        conn, table = self.get_model(domain, Settings)
        for row in table.filter(dict(key=key)).limit(1).run(conn):
            if parser is not None:
                return parser(row['value'])
            else:
                return row['value']
        return fail

    def _fetch_record(self, domain, model, **terms):
        conn, table = self.get_model(domain, model)
        for row in table.filter(terms).limit(1).run(conn):
            return Record(model, row['uuid'], vsn=row['vsn'],
                          txn=row['txn'], **loads(row['data']))

    def fetch_profile(self, domain, identity, **extra):
        return self._fetch_record(domain, Profile, username=identity, **extra)

    def prelogin_count(self, domain, origin):
        conn, table = self.get_model(domain, Prelogin)
        return table.count(dict(origin=origin)).run(conn)

    def fetch_prelogin(self, origin, domain, sid):
        conn, table = self.get_model(domain, Prelogin)
        try:
            enc_sid = b64encode(sid)
            rs = table.get(enc_sid).delete(return_vals=True).run(conn)
            prelogin = Hict(rs['old_val'])
            assert b64decode(prelogin.sid) == sid
            prelogin.update(**loads(prelogin.pop('data')))
            return prelogin
        except:
            pass

    def store_prelogin(self, origin, domain, sid, identity, expires, **data):
        row = dict(origin=origin, identity=identity, expires=expires,
                   sid=b64encode(sid), data=dumps(data))
        conn, table = self.get_model(domain, Prelogin)
        table.insert(row).run(conn)

    def load_state(self, domain, sid):
        conn, table = self.get_model(domain, State)
        row = table.get(b64encode(sid)).run(conn)
        if row:
            assert sid == b64decode(row.pop('sid'))
            profile = self._fetch_record(domain, Profile, uuid=row.pop('profile'))  # noqa
            state = Hict(sid=sid, **loads(row.pop('data'), dict_type=Hict))
            state.update(profile=profile, **row)
            return state

    def store_state(self, state):
        profile = state.profile
        data = dict(skey=state.skey, ckey=state.ckey, sseed=state.sseed,
                    vkey=state.vkey, pads=state.pads, data=state.data,
                    role=state.role, impersonated_by=state.impersonated_by)
        row = dict(sid=b64encode(state.sid), identity=profile.username,
                   profile=str(profile._uuid), origin=state.origin,
                   expires=state.expires, data=dumps(data))
        conn, table = self.get_model(state.domain, State)
        try:
            assert table.update(row).run(conn)['replaced'] == 1
        except:
            assert table.insert(row).run(conn)['inserted'] == 1

    def destroy_state(self, state):
        conn, table = self.get_model(state.domain, State)
        table.get(b64encode(state.sid)).delete().run(conn)

    def active_user_sessions(self, domain, identity):
        conn, table = self.get_model(domain, State)
        qry = (rtdb.row['identity'] == identity) \
            & (rtdb.row['expires'] > timestamp())
        return table.count(qry).run(conn)

    def destroy_sessions(self, domain, identity):
        conn, table = self.get_model(domain, State)
        table.filter(dict(identity=identity)).delete().run(conn)

    def incr_abuse(self, origin, domain, incr=1, abuse=None):
        data = dumps(abuse) if abuse is not None else abuse
        row = dict(origin=origin, level=incr, tstamp=timestamp(), data=data)
        conn, table = self.get_model(domain, Abuse)
        table.insert(row).run(conn)

    def abuse_level_by_origin(self, origin, domain):
        conn, table = self.get_model(domain, Abuse)
        watch_period = timestamp() - hooks.abuse_level_watch_period(domain)
        qry = (rtdb.row['origin'] == origin) \
            & (rtdb.row['tstamp'] >= watch_period)
        return sum(doc['level'] for doc in table.filter(qry).run(conn))

    def session_timeout(self, domain):
        return self._fetch_setting(domain, 'session_timeout')

    def max_prelogin_count(self, domain):
        return self._fetch_setting(domain, 'max_prelogin_count')

    def abuse_level_watch_period(self, domain):
        return self._fetch_setting(domain, 'abuse_level_watch_period')

    def max_origin_abuse_level(self, domain):
        return self._fetch_setting(domain, 'max_origin_abuse_level')

    def max_active_user_sessions(self, domain, identity):
        return self._fetch_setting(domain, 'max_active_user_sessions')

    def has_permission(self, state, call):
        return True



class FaneryException(Exception):
    pass


class MultipleDefine(FaneryException):
    pass


class UndefinedOutputFormatter(FaneryException):
    pass


class InvalidOutputFormatter(FaneryException):
    pass


class CallPadsLimitExceeded(FaneryException):
    pass


class InvalidCall(FaneryException):
    pass


class NotFound(FaneryException):
    pass


class Abusive(FaneryException):
    pass


class Forbidden(FaneryException):
    pass


class Unauthorized(FaneryException):
    pass


class RequireSSL(FaneryException):
    pass


class CallPadTimeout(FaneryException):
    pass


class ExpiredSession(FaneryException):
    pass


class UnknownSession(FaneryException):
    pass


class InvalidCredential(FaneryException):
    pass


class MaxActiveSessions(FaneryException):
    pass


class DuplicatedRecordID(FaneryException):
    pass


class RecordValidationError(FaneryException):
    pass


class RecordVersionMismatch(FaneryException):
    pass


class RecordNotFound(FaneryException):
    pass


class UnstoredRecord(FaneryException):
    pass


class DeletedRecord(FaneryException):
    pass


class MultipleRecordsFound(FaneryException):
    pass


class RecordMarkedForDeletion(FaneryException):
    pass


class RecordMarkedForInsertion(FaneryException):
    pass



from nacl.public import PrivateKey, PublicKey, Box
from nacl.signing import SigningKey, VerifyKey
from nacl.utils import random
from nacl.hash import sha256


def nacl_sha256(text):
    return sha256(text).decode('hex')


def nacl_random(size):
    return random(size)


def nacl_nonce():
    return random(Box.NONCE_SIZE)


def nacl_box(pub_key, secret=None):
    pri_key = PrivateKey(secret) if secret else PrivateKey.generate()
    pub_key = PublicKey(pub_key)
    return Box(pri_key, pub_key), pri_key, pub_key


def nacl_sign(msg, seed=None):
    key = SigningKey(seed) if seed else SigningKey.generate()
    return key.sign(msg), key.verify_key


def nacl_verify(signed, key):
    return VerifyKey(key).verify(signed)



from setuptools import setup, find_packages

exec(open('fanery/_version.py'))
setup(
    name='Fanery',
    version=__version__, # noqa
    author='Marco Caramma',
    author_email='marco@globalsoftwaresecurity.com',
    url='https://bitbucket.org/mcaramma/fanery/',
    license='ISC',
    description='Application development framework',
    long_description=open('README.rst').read(),
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'webob',
        'PyNaCl',
        'cxor',
        'ujson',
        'scrypt',
        'bsdiff4',
        'ciso8601',
        'python-libuuid',
        'msgpack-python',
        ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Operating System :: POSIX :: Linux',
        'Operating System :: POSIX :: BSD',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Internet :: WWW/HTTP :: Session',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Security',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



from _term import *         # noqa
from _random import *       # noqa
from _version import *      # noqa
from _profile import *      # noqa
from _service import *      # noqa
from _client import *       # noqa
from _state import *        # noqa
from _store import *        # noqa
import _dbm as dbm          # noqa
import _exc as exc          # noqa
import _wsgi as wsgi        # noqa
import _admin as admin      # noqa
import _server as server    # noqa
import _config as config    # noqa
